# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common / shared code for handling authentication against OpenStack identity
service (Keystone).
"""

import datetime
from collections import namedtuple

from libcloud.utils.py3 import httplib
from libcloud.common.base import Response, ConnectionUserAndKey, CertificateConnection
from libcloud.compute.types import LibcloudError, InvalidCredsError, MalformedResponseError
from libcloud.utils.iso8601 import parse_date

try:
    import simplejson as json
except ImportError:
    import json  # type: ignore

AUTH_API_VERSION = "1.1"
AUTH_TOKEN_HEADER = "X-Auth-Token"

# Auth versions which contain token expiration information.
AUTH_VERSIONS_WITH_EXPIRES = [
    "1.1",
    "2.0",
    "2.0_apikey",
    "2.0_password",
    "2.0_voms",
    "3.0",
    "3.x_password",
    "3.x_appcred",
    "3.x_oidc_access_token",
]

# How many seconds to subtract from the auth token expiration time before
# testing if the token is still valid.
# The time is subtracted to account for the HTTP request latency and prevent
# user from getting "InvalidCredsError" if token is about to expire.
AUTH_TOKEN_EXPIRES_GRACE_SECONDS = 5


__all__ = [
    "OpenStackAuthenticationCache",
    "OpenStackAuthenticationCacheKey",
    "OpenStackAuthenticationContext",
    "OpenStackIdentityVersion",
    "OpenStackIdentityDomain",
    "OpenStackIdentityProject",
    "OpenStackIdentityUser",
    "OpenStackIdentityRole",
    "OpenStackServiceCatalog",
    "OpenStackServiceCatalogEntry",
    "OpenStackServiceCatalogEntryEndpoint",
    "OpenStackIdentityEndpointType",
    "OpenStackIdentityConnection",
    "OpenStackIdentity_1_0_Connection",
    "OpenStackIdentity_1_1_Connection",
    "OpenStackIdentity_2_0_Connection",
    "OpenStackIdentity_2_0_Connection_VOMS",
    "OpenStackIdentity_3_0_Connection",
    "OpenStackIdentity_3_0_Connection_AppCred",
    "OpenStackIdentity_3_0_Connection_OIDC_access_token",
    "get_class_for_auth_version",
]


class OpenStackAuthenticationCache:
    """
    Base class for external OpenStack authentication caches.

    Authentication tokens are always cached in memory in
    :class:`OpenStackIdentityConnection`.auth_token and related fields.  These
    tokens are lost when the driver is garbage collected.  To share tokens
    among multiple drivers, processes, or systems, use an
    :class:`OpenStackAuthenticationCache` in
    OpenStackIdentityConnection.auth_cache.

    Cache implementers should inherit this class and define the methods below.
    """

    def get(self, key):
        """
        Get an authentication context from the cache.

        :param key: Key to fetch.
        :type key: :class:`.OpenStackAuthenticationCacheKey`

        :return: The cached context for the given key, if present; None if not.
        :rtype: :class:`OpenStackAuthenticationContext`
        """
        raise NotImplementedError

    def put(self, key, context):
        """
        Put an authentication context into the cache.

        :param key: Key where the context will be stored.
        :type key: :class:`.OpenStackAuthenticationCacheKey`

        :param context: The context to cache.
        :type context: :class:`.OpenStackAuthenticationContext`
        """
        raise NotImplementedError

    def clear(self, key):
        """
        Clear an authentication context from the cache.

        :param key: Key to clear.
        :type key: :class:`.OpenStackAuthenticationCacheKey`
        """
        raise NotImplementedError


OpenStackAuthenticationCacheKey = namedtuple(
    "OpenStackAuthenticationCacheKey",
    [
        "auth_url",
        "user_id",
        "token_scope",
        "tenant_name",
        "domain_name",
        "tenant_domain_id",
    ],
)


class OpenStackAuthenticationContext:
    """
    An authentication token and related context.
    """

    def __init__(self, token, expiration=None, user=None, roles=None, urls=None):
        self.token = token
        self.expiration = expiration
        self.user = user
        self.roles = roles
        self.urls = urls


class OpenStackIdentityEndpointType:
    """
    Enum class for openstack identity endpoint type.
    """

    INTERNAL = "internal"
    EXTERNAL = "external"
    ADMIN = "admin"


class OpenStackIdentityTokenScope:
    """
    Enum class for openstack identity token scope.
    """

    PROJECT = "project"
    DOMAIN = "domain"
    UNSCOPED = "unscoped"


class OpenStackIdentityVersion:
    def __init__(self, version, status, updated, url):
        self.version = version
        self.status = status
        self.updated = updated
        self.url = url

    def __repr__(self):
        return "<OpenStackIdentityVersion version=%s, status=%s, " "updated=%s, url=%s>" % (
            self.version,
            self.status,
            self.updated,
            self.url,
        )


class OpenStackIdentityDomain:
    def __init__(self, id, name, enabled):
        self.id = id
        self.name = name
        self.enabled = enabled

    def __repr__(self):
        return "<OpenStackIdentityDomain id={}, name={}, enabled={}>".format(
            self.id,
            self.name,
            self.enabled,
        )


class OpenStackIdentityProject:
    def __init__(self, id, name, description, enabled, domain_id=None):
        self.id = id
        self.name = name
        self.description = description
        self.enabled = enabled
        self.domain_id = domain_id

    def __repr__(self):
        return "<OpenStackIdentityProject id=%s, domain_id=%s, name=%s, " "enabled=%s>" % (
            self.id,
            self.domain_id,
            self.name,
            self.enabled,
        )


class OpenStackIdentityRole:
    def __init__(self, id, name, description, enabled):
        self.id = id
        self.name = name
        self.description = description
        self.enabled = enabled

    def __repr__(self):
        return "<OpenStackIdentityRole id=%s, name=%s, description=%s, " "enabled=%s>" % (
            self.id,
            self.name,
            self.description,
            self.enabled,
        )


class OpenStackIdentityUser:
    def __init__(self, id, domain_id, name, email, description, enabled):
        self.id = id
        self.domain_id = domain_id
        self.name = name
        self.email = email
        self.description = description
        self.enabled = enabled

    def __repr__(self):
        return "<OpenStackIdentityUser id=%s, domain_id=%s, name=%s, " "email=%s, enabled=%s>" % (
            self.id,
            self.domain_id,
            self.name,
            self.email,
            self.enabled,
        )


class OpenStackServiceCatalog:
    """
    http://docs.openstack.org/api/openstack-identity-service/2.0/content/

    This class should be instantiated with the contents of the
    'serviceCatalog' in the auth response. This will do the work of figuring
    out which services actually exist in the catalog as well as split them up
    by type, name, and region if available
    """

    _auth_version = None
    _service_catalog = None

    def __init__(self, service_catalog, auth_version=AUTH_API_VERSION):
        self._auth_version = auth_version

        # Check this way because there are a couple of different 2.0_*
        # auth types.
        if "3.x" in self._auth_version:
            entries = self._parse_service_catalog_auth_v3(service_catalog=service_catalog)
        elif "2.0" in self._auth_version:
            entries = self._parse_service_catalog_auth_v2(service_catalog=service_catalog)
        elif ("1.1" in self._auth_version) or ("1.0" in self._auth_version):
            entries = self._parse_service_catalog_auth_v1(service_catalog=service_catalog)
        else:
            raise LibcloudError('auth version "%s" not supported' % (self._auth_version))

        # Force consistent ordering by sorting the entries
        entries = sorted(entries, key=lambda x: x.service_type + (x.service_name or ""))
        self._entries = entries  # stories all the service catalog entries

    def get_entries(self):
        """
        Return all the entries for this service catalog.

        :rtype: ``list`` of :class:`.OpenStackServiceCatalogEntry`
        """
        return self._entries

    def get_catalog(self):
        """
        Deprecated in the favor of ``get_entries`` method.
        """
        return self.get_entries()

    def get_public_urls(self, service_type=None, name=None):
        """
        Retrieve all the available public (external) URLs for the provided
        service type and name.
        """
        endpoints = self.get_endpoints(service_type=service_type, name=name)

        result = []
        for endpoint in endpoints:
            endpoint_type = endpoint.endpoint_type
            if endpoint_type == OpenStackIdentityEndpointType.EXTERNAL:
                result.append(endpoint.url)

        return result

    def get_endpoints(self, service_type=None, name=None):
        """
        Retrieve all the endpoints for the provided service type and name.

        :rtype: ``list`` of :class:`.OpenStackServiceCatalogEntryEndpoint`
        """
        endpoints = []

        for entry in self._entries:
            # Note: "if XXX and YYY != XXX" comparison is used to support
            # partial lookups.
            # This allows user to pass in only one argument to the method (only
            # service_type or name), both of them or neither.
            if service_type and entry.service_type != service_type:
                continue

            if name and entry.service_name != name:
                continue

            for endpoint in entry.endpoints:
                endpoints.append(endpoint)

        return endpoints

    def get_endpoint(
        self,
        service_type=None,
        name=None,
        region=None,
        endpoint_type=OpenStackIdentityEndpointType.EXTERNAL,
    ):
        """
        Retrieve a single endpoint using the provided criteria.

        Note: If no or more than one matching endpoint is found, an exception
        is thrown.
        """
        endpoints = []

        for entry in self._entries:
            if service_type and entry.service_type != service_type:
                continue

            if name and entry.service_name != name:
                continue

            for endpoint in entry.endpoints:
                if region and endpoint.region != region:
                    continue

                if endpoint_type and endpoint.endpoint_type != endpoint_type:
                    continue

                endpoints.append(endpoint)

        if len(endpoints) == 1:
            return endpoints[0]
        elif len(endpoints) > 1:
            raise ValueError("Found more than 1 matching endpoint")
        else:
            raise LibcloudError("Could not find specified endpoint")

    def get_regions(self, service_type=None):
        """
        Retrieve a list of all the available regions.

        :param service_type: If specified, only return regions for this
                             service type.
        :type service_type: ``str``

        :rtype: ``list`` of ``str``
        """
        regions = set()

        for entry in self._entries:
            if service_type and entry.service_type != service_type:
                continue

            for endpoint in entry.endpoints:
                if endpoint.region:
                    regions.add(endpoint.region)

        return sorted(list(regions))

    def get_service_types(self, region=None):
        """
        Retrieve all the available service types.

        :param region: Optional region to retrieve service types for.
        :type region: ``str``

        :rtype: ``list`` of ``str``
        """
        service_types = set()

        for entry in self._entries:
            include = True

            for endpoint in entry.endpoints:
                if region and endpoint.region != region:
                    include = False
                    break

            if include:
                service_types.add(entry.service_type)

        return sorted(list(service_types))

    def get_service_names(self, service_type=None, region=None):
        """
        Retrieve list of service names that match service type and region.

        :type service_type: ``str``
        :type region: ``str``

        :rtype: ``list`` of ``str``
        """
        names = set()

        if "2.0" not in self._auth_version:
            raise ValueError("Unsupported version: %s" % (self._auth_version))

        for entry in self._entries:
            if service_type and entry.service_type != service_type:
                continue

            include = True
            for endpoint in entry.endpoints:
                if region and endpoint.region != region:
                    include = False
                    break

            if include and entry.service_name:
                names.add(entry.service_name)

        return sorted(list(names))

    def _parse_service_catalog_auth_v1(self, service_catalog):
        entries = []

        for service, endpoints in service_catalog.items():
            entry_endpoints = []
            for endpoint in endpoints:
                region = endpoint.get("region", None)

                public_url = endpoint.get("publicURL", None)
                private_url = endpoint.get("internalURL", None)

                if public_url:
                    entry_endpoint = OpenStackServiceCatalogEntryEndpoint(
                        region=region,
                        url=public_url,
                        endpoint_type=OpenStackIdentityEndpointType.EXTERNAL,
                    )
                    entry_endpoints.append(entry_endpoint)

                if private_url:
                    entry_endpoint = OpenStackServiceCatalogEntryEndpoint(
                        region=region,
                        url=private_url,
                        endpoint_type=OpenStackIdentityEndpointType.INTERNAL,
                    )
                    entry_endpoints.append(entry_endpoint)

            entry = OpenStackServiceCatalogEntry(service_type=service, endpoints=entry_endpoints)
            entries.append(entry)

        return entries

    def _parse_service_catalog_auth_v2(self, service_catalog):
        entries = []

        for service in service_catalog:
            service_type = service["type"]
            service_name = service.get("name", None)

            entry_endpoints = []
            for endpoint in service.get("endpoints", []):
                region = endpoint.get("region", None)

                public_url = endpoint.get("publicURL", None)
                private_url = endpoint.get("internalURL", None)

                if public_url:
                    entry_endpoint = OpenStackServiceCatalogEntryEndpoint(
                        region=region,
                        url=public_url,
                        endpoint_type=OpenStackIdentityEndpointType.EXTERNAL,
                    )
                    entry_endpoints.append(entry_endpoint)

                if private_url:
                    entry_endpoint = OpenStackServiceCatalogEntryEndpoint(
                        region=region,
                        url=private_url,
                        endpoint_type=OpenStackIdentityEndpointType.INTERNAL,
                    )
                    entry_endpoints.append(entry_endpoint)

            entry = OpenStackServiceCatalogEntry(
                service_type=service_type,
                endpoints=entry_endpoints,
                service_name=service_name,
            )
            entries.append(entry)

        return entries

    def _parse_service_catalog_auth_v3(self, service_catalog):
        entries = []

        for item in service_catalog:
            service_type = item["type"]
            service_name = item.get("name", None)

            entry_endpoints = []
            for endpoint in item["endpoints"]:
                region = endpoint.get("region", None)
                url = endpoint["url"]
                endpoint_type = endpoint["interface"]

                if endpoint_type == "internal":
                    endpoint_type = OpenStackIdentityEndpointType.INTERNAL
                elif endpoint_type == "public":
                    endpoint_type = OpenStackIdentityEndpointType.EXTERNAL
                elif endpoint_type == "admin":
                    endpoint_type = OpenStackIdentityEndpointType.ADMIN

                entry_endpoint = OpenStackServiceCatalogEntryEndpoint(
                    region=region, url=url, endpoint_type=endpoint_type
                )
                entry_endpoints.append(entry_endpoint)

            entry = OpenStackServiceCatalogEntry(
                service_type=service_type,
                service_name=service_name,
                endpoints=entry_endpoints,
            )
            entries.append(entry)

        return entries


class OpenStackServiceCatalogEntry:
    def __init__(self, service_type, endpoints=None, service_name=None):
        """
        :param service_type: Service type.
        :type service_type: ``str``

        :param endpoints: Endpoints belonging to this entry.
        :type endpoints: ``list``

        :param service_name: Optional service name.
        :type service_name: ``str``
        """
        self.service_type = service_type
        self.endpoints = endpoints or []
        self.service_name = service_name

        # For consistency, sort the endpoints
        self.endpoints = sorted(self.endpoints, key=lambda x: x.url or "")

    def __eq__(self, other):
        return (
            self.service_type == other.service_type
            and self.endpoints == other.endpoints
            and other.service_name == self.service_name
        )

    def __ne__(self, other):
        return not self.__eq__(other=other)

    def __repr__(self):
        return "<OpenStackServiceCatalogEntry service_type=%s, " "service_name=%s, endpoints=%s" % (
            self.service_type,
            self.service_name,
            repr(self.endpoints),
        )


class OpenStackServiceCatalogEntryEndpoint:
    VALID_ENDPOINT_TYPES = [
        OpenStackIdentityEndpointType.INTERNAL,
        OpenStackIdentityEndpointType.EXTERNAL,
        OpenStackIdentityEndpointType.ADMIN,
    ]

    def __init__(self, region, url, endpoint_type="external"):
        """
        :param region: Endpoint region.
        :type region: ``str``

        :param url: Endpoint URL.
        :type url: ``str``

        :param endpoint_type: Endpoint type (external / internal / admin).
        :type endpoint_type: ``str``
        """
        if endpoint_type not in self.VALID_ENDPOINT_TYPES:
            raise ValueError("Invalid type: %s" % (endpoint_type))

        # TODO: Normalize / lowercase all the region names
        self.region = region
        self.url = url
        self.endpoint_type = endpoint_type

    def __eq__(self, other):
        return (
            self.region == other.region
            and self.url == other.url
            and self.endpoint_type == other.endpoint_type
        )

    def __ne__(self, other):
        return not self.__eq__(other=other)

    def __repr__(self):
        return "<OpenStackServiceCatalogEntryEndpoint region=%s, url=%s, " "type=%s" % (
            self.region,
            self.url,
            self.endpoint_type,
        )


class OpenStackAuthResponse(Response):
    def success(self):
        return self.status in [
            httplib.OK,
            httplib.CREATED,
            httplib.ACCEPTED,
            httplib.NO_CONTENT,
            httplib.MULTIPLE_CHOICES,
            httplib.UNAUTHORIZED,
            httplib.INTERNAL_SERVER_ERROR,
        ]

    def parse_body(self):
        if not self.body:
            return None

        if "content-type" in self.headers:
            key = "content-type"
        elif "Content-Type" in self.headers:
            key = "Content-Type"
        else:
            raise LibcloudError("Missing content-type header", driver=OpenStackIdentityConnection)

        content_type = self.headers[key]
        if content_type.find(";") != -1:
            content_type = content_type.split(";")[0]

        if content_type == "application/json":
            try:
                data = json.loads(self.body)
            except Exception:
                driver = OpenStackIdentityConnection
                raise MalformedResponseError("Failed to parse JSON", body=self.body, driver=driver)
        elif content_type == "text/plain":
            data = self.body
        else:
            data = self.body

        return data


class OpenStackIdentityConnection(ConnectionUserAndKey):
    """
    Base identity connection class which contains common / shared logic.

    Note: This class shouldn't be instantiated directly.
    """

    responseCls = OpenStackAuthResponse
    timeout = None
    auth_version = None  # type: str

    def __init__(
        self,
        auth_url,
        user_id,
        key,
        tenant_name=None,
        tenant_domain_id="default",
        domain_name="Default",
        token_scope=OpenStackIdentityTokenScope.PROJECT,
        timeout=None,
        proxy_url=None,
        parent_conn=None,
        auth_cache=None,
    ):
        super().__init__(
            user_id=user_id, key=key, url=auth_url, timeout=timeout, proxy_url=proxy_url
        )

        self.parent_conn = parent_conn

        # enable tests to use the same mock connection classes.
        if parent_conn:
            self.conn_class = parent_conn.conn_class
            self.driver = parent_conn.driver
        else:
            self.driver = None

        self.auth_url = auth_url
        self.tenant_name = tenant_name
        self.domain_name = domain_name
        self.tenant_domain_id = tenant_domain_id
        self.token_scope = token_scope
        self.timeout = timeout
        self.auth_cache = auth_cache

        self.urls = {}
        self.auth_token = None
        self.auth_token_expires = None
        self.auth_user_info = None
        self.auth_user_roles = None

    def authenticated_request(
        self, action, params=None, data=None, headers=None, method="GET", raw=False
    ):
        """
        Perform an authenticated request against the identity API.
        """
        if not self.auth_token:
            raise ValueError("Need to be authenticated to perform this request")

        headers = headers or {}
        headers[AUTH_TOKEN_HEADER] = self.auth_token

        response = self.request(
            action=action,
            params=params,
            data=data,
            headers=headers,
            method=method,
            raw=raw,
        )
        # Evict cached auth token if we receive Unauthorized while using it
        if response.status == httplib.UNAUTHORIZED:
            self.clear_cached_auth_context()
        return response

    def morph_action_hook(self, action):
        _, _, _, request_path = self._tuple_from_url(self.auth_url)

        if request_path == "":
            # No path is provided in the auth_url, use action passed to this
            # method.
            return action

        return super().morph_action_hook(action=action)

    def add_default_headers(self, headers):
        headers["Accept"] = "application/json"
        headers["Content-Type"] = "application/json; charset=UTF-8"
        return headers

    def is_token_valid(self):
        """
        Return True if the current auth token is already cached and hasn't
        expired yet.

        :return: ``True`` if the token is still valid, ``False`` otherwise.
        :rtype: ``bool``
        """
        if not self.auth_token:
            return False

        if not self.auth_token_expires:
            return False

        expires = self.auth_token_expires - datetime.timedelta(
            seconds=AUTH_TOKEN_EXPIRES_GRACE_SECONDS
        )

        time_tuple_expires = expires.utctimetuple()
        time_tuple_now = datetime.datetime.utcnow().utctimetuple()

        if time_tuple_now < time_tuple_expires:
            return True

        return False

    def authenticate(self, force=False):
        """
        Authenticate against the identity API.

        :param force: Forcefully update the token even if it's already cached
                      and still valid.
        :type force: ``bool``
        """
        raise NotImplementedError("authenticate not implemented")

    def clear_cached_auth_context(self):
        """
        Clear the cached authentication context.

        The context is cleared from fields on this connection and from the
        external cache, if one is configured.
        """
        self.auth_token = None
        self.auth_token_expires = None
        self.auth_user_info = None
        self.auth_user_roles = None
        self.urls = {}

        if self.auth_cache is not None:
            self.auth_cache.clear(self._cache_key)

    def list_supported_versions(self):
        """
        Retrieve a list of all the identity versions which are supported by
        this installation.

        :rtype: ``list`` of :class:`.OpenStackIdentityVersion`
        """
        response = self.request("/", method="GET")
        result = self._to_versions(data=response.object["versions"]["values"])
        result = sorted(result, key=lambda x: x.version)
        return result

    def _to_versions(self, data):
        result = []
        for item in data:
            version = self._to_version(data=item)
            result.append(version)

        return result

    def _to_version(self, data):
        try:
            updated = parse_date(data["updated"])
        except Exception:
            updated = None

        try:
            url = data["links"][0]["href"]
        except IndexError:
            url = None

        version = OpenStackIdentityVersion(
            version=data["id"], status=data["status"], updated=updated, url=url
        )
        return version

    def _is_authentication_needed(self, force=False):
        """
        Determine if the authentication is needed or if the existing token (if
        any exists) is still valid.
        """
        if force:
            return True

        if self.auth_version not in AUTH_VERSIONS_WITH_EXPIRES:
            return True

        if self.is_token_valid():
            return False

        # See if there's a new token in the cache
        self._load_auth_context_from_cache()

        # If there was a token in the cache, it is now stored in our local
        # auth_token and related fields.  Ensure it is still valid.
        if self.is_token_valid():
            return False

        return True

    def _to_projects(self, data):
        result = []
        for item in data:
            project = self._to_project(data=item)
            result.append(project)

        return result

    def _to_project(self, data):
        project = OpenStackIdentityProject(
            id=data["id"],
            name=data["name"],
            description=data["description"],
            enabled=data["enabled"],
            domain_id=data.get("domain_id", None),
        )
        return project

    @property
    def _cache_key(self):
        """
        The key where this connection's authentication context will be cached.

        :rtype: :class:`OpenStackAuthenticationCacheKey`
        """
        return OpenStackAuthenticationCacheKey(
            self.auth_url,
            self.user_id,
            self.token_scope,
            self.tenant_name,
            self.domain_name,
            self.tenant_domain_id,
        )

    def _cache_auth_context(self, context):
        """
        Store an authentication context in memory and the cache.

        :param context: Authentication context to cache.
        :type key: :class:`.OpenStackAuthenticationContext`
        """
        self.urls = context.urls
        self.auth_token = context.token
        self.auth_token_expires = context.expiration
        self.auth_user_info = context.user
        self.auth_user_roles = context.roles

        if self.auth_cache is not None:
            self.auth_cache.put(self._cache_key, context)

    def _load_auth_context_from_cache(self):
        """
        Fetch an authentication context for this connection from the cache.

        :rtype: :class:`OpenStackAuthenticationContext`
        """
        if self.auth_cache is None:
            return None

        context = self.auth_cache.get(self._cache_key)
        if context is None:
            return None

        self.urls = context.urls
        self.auth_token = context.token
        self.auth_token_expires = context.expiration
        self.auth_user_info = context.user
        self.auth_user_roles = context.roles
        return context


class OpenStackIdentity_1_0_Connection(OpenStackIdentityConnection):
    """
    Connection class for Keystone API v1.0.
    """

    responseCls = OpenStackAuthResponse
    name = "OpenStack Identity API v1.0"
    auth_version = "1.0"

    def authenticate(self, force=False):
        if not self._is_authentication_needed(force=force):
            return self

        headers = {
            "X-Auth-User": self.user_id,
            "X-Auth-Key": self.key,
        }

        resp = self.request("/v1.0", headers=headers, method="GET")

        if resp.status == httplib.UNAUTHORIZED:
            # HTTP UNAUTHORIZED (401): auth failed
            raise InvalidCredsError()
        elif resp.status not in [httplib.NO_CONTENT, httplib.OK]:
            body = "code: {} body:{} headers:{}".format(
                resp.status,
                resp.body,
                resp.headers,
            )
            raise MalformedResponseError("Malformed response", body=body, driver=self.driver)
        else:
            headers = resp.headers
            # emulate the auth 1.1 URL list
            self.urls = {}
            self.urls["cloudServers"] = [
                {"publicURL": headers.get("x-server-management-url", None)}
            ]
            self.urls["cloudFilesCDN"] = [{"publicURL": headers.get("x-cdn-management-url", None)}]
            self.urls["cloudFiles"] = [{"publicURL": headers.get("x-storage-url", None)}]
            self.auth_token = headers.get("x-auth-token", None)
            self.auth_user_info = None

            if not self.auth_token:
                raise MalformedResponseError("Missing X-Auth-Token in" " response headers")

        return self


class OpenStackIdentity_1_1_Connection(OpenStackIdentityConnection):
    """
    Connection class for Keystone API v1.1.
    """

    responseCls = OpenStackAuthResponse
    name = "OpenStack Identity API v1.1"
    auth_version = "1.1"

    def authenticate(self, force=False):
        if not self._is_authentication_needed(force=force):
            return self

        reqbody = json.dumps({"credentials": {"username": self.user_id, "key": self.key}})
        resp = self.request("/v1.1/auth", data=reqbody, headers={}, method="POST")

        if resp.status == httplib.UNAUTHORIZED:
            # HTTP UNAUTHORIZED (401): auth failed
            raise InvalidCredsError()
        elif resp.status != httplib.OK:
            body = "code: {} body:{}".format(resp.status, resp.body)
            raise MalformedResponseError("Malformed response", body=body, driver=self.driver)
        else:
            try:
                body = json.loads(resp.body)
            except Exception as e:
                raise MalformedResponseError("Failed to parse JSON", e)

            try:
                expires = body["auth"]["token"]["expires"]
                self._cache_auth_context(
                    OpenStackAuthenticationContext(
                        body["auth"]["token"]["id"],
                        expiration=parse_date(expires),
                        urls=body["auth"]["serviceCatalog"],
                    )
                )
            except KeyError as e:
                raise MalformedResponseError(
                    "Auth JSON response is \
                                             missing required elements",
                    e,
                )

        return self


class OpenStackIdentity_2_0_Connection(OpenStackIdentityConnection):
    """
    Connection class for Keystone API v2.0.
    """

    responseCls = OpenStackAuthResponse
    name = "OpenStack Identity API v1.0"
    auth_version = "2.0"

    def authenticate(self, auth_type="api_key", force=False):
        if not self._is_authentication_needed(force=force):
            return self

        if auth_type == "api_key":
            return self._authenticate_2_0_with_api_key()
        elif auth_type == "password":
            return self._authenticate_2_0_with_password()
        else:
            raise ValueError("Invalid value for auth_type argument")

    def _authenticate_2_0_with_api_key(self):
        # API Key based authentication uses the RAX-KSKEY extension.
        # http://s.apache.org/oAi
        data = {
            "auth": {
                "RAX-KSKEY:apiKeyCredentials": {
                    "username": self.user_id,
                    "apiKey": self.key,
                }
            }
        }
        if self.tenant_name:
            data["auth"]["tenantName"] = self.tenant_name
        reqbody = json.dumps(data)
        return self._authenticate_2_0_with_body(reqbody)

    def _authenticate_2_0_with_password(self):
        # Password based authentication is the only 'core' authentication
        # method in Keystone at this time.
        # 'keystone' - http://s.apache.org/e8h
        data = {"auth": {"passwordCredentials": {"username": self.user_id, "password": self.key}}}
        if self.tenant_name:
            data["auth"]["tenantName"] = self.tenant_name
        reqbody = json.dumps(data)
        return self._authenticate_2_0_with_body(reqbody)

    def _authenticate_2_0_with_body(self, reqbody):
        resp = self.request(
            "/v2.0/tokens",
            data=reqbody,
            headers={"Content-Type": "application/json"},
            method="POST",
        )

        if resp.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError()
        elif resp.status not in [httplib.OK, httplib.NON_AUTHORITATIVE_INFORMATION]:
            body = "code: {} body: {}".format(resp.status, resp.body)
            raise MalformedResponseError("Malformed response", body=body, driver=self.driver)
        else:
            body = resp.object

            try:
                access = body["access"]
                expires = access["token"]["expires"]
                self._cache_auth_context(
                    OpenStackAuthenticationContext(
                        access["token"]["id"],
                        expiration=parse_date(expires),
                        urls=access["serviceCatalog"],
                        user=access.get("user", {}),
                    )
                )
            except KeyError as e:
                raise MalformedResponseError(
                    "Auth JSON response is \
                                             missing required elements",
                    e,
                )

        return self

    def list_projects(self):
        response = self.authenticated_request("/v2.0/tenants", method="GET")
        result = self._to_projects(data=response.object["tenants"])
        return result

    def list_tenants(self):
        return self.list_projects()


class OpenStackIdentity_3_0_Connection(OpenStackIdentityConnection):
    """
    Connection class for Keystone API v3.x.
    """

    responseCls = OpenStackAuthResponse
    name = "OpenStack Identity API v3.x"
    auth_version = "3.0"

    VALID_TOKEN_SCOPES = [
        OpenStackIdentityTokenScope.PROJECT,
        OpenStackIdentityTokenScope.DOMAIN,
        OpenStackIdentityTokenScope.UNSCOPED,
    ]

    def __init__(
        self,
        auth_url,
        user_id,
        key,
        tenant_name=None,
        domain_name="Default",
        tenant_domain_id="default",
        token_scope=OpenStackIdentityTokenScope.PROJECT,
        timeout=None,
        proxy_url=None,
        parent_conn=None,
        auth_cache=None,
    ):
        """
        :param tenant_name: Name of the project this user belongs to. Note:
                            When token_scope is set to project, this argument
                            control to which project to scope the token to.
        :type tenant_name: ``str``

        :param domain_name: Domain the user belongs to. Note: When token_scope
                            is set to token, this argument controls to which
                            domain to scope the token to.
        :type domain_name: ``str``

        :param token_scope: Whether to scope a token to a "project", a
                            "domain" or "unscoped"
        :type token_scope: ``str``

        :param auth_cache: Where to cache authentication tokens.
        :type auth_cache: :class:`OpenStackAuthenticationCache`
        """
        super().__init__(
            auth_url=auth_url,
            user_id=user_id,
            key=key,
            tenant_name=tenant_name,
            domain_name=domain_name,
            tenant_domain_id=tenant_domain_id,
            token_scope=token_scope,
            timeout=timeout,
            proxy_url=proxy_url,
            parent_conn=parent_conn,
            auth_cache=auth_cache,
        )

        if self.token_scope not in self.VALID_TOKEN_SCOPES:
            raise ValueError('Invalid value for "token_scope" argument: %s' % (self.token_scope))

        if self.token_scope == OpenStackIdentityTokenScope.PROJECT and (
            not self.tenant_name or not self.domain_name
        ):
            raise ValueError("Must provide tenant_name and domain_name " "argument")
        elif self.token_scope == OpenStackIdentityTokenScope.DOMAIN and not self.domain_name:
            raise ValueError("Must provide domain_name argument")

    def authenticate(self, force=False):
        """
        Perform authentication.
        """
        if not self._is_authentication_needed(force=force):
            return self

        data = self._get_auth_data()
        data = json.dumps(data)
        response = self.request(
            "/v3/auth/tokens",
            data=data,
            headers={"Content-Type": "application/json"},
            method="POST",
        )
        self._parse_token_response(response, cache_it=True)
        return self

    def list_domains(self):
        """
        List the available domains.

        :rtype: ``list`` of :class:`OpenStackIdentityDomain`
        """
        response = self.authenticated_request("/v3/domains", method="GET")
        result = self._to_domains(data=response.object["domains"])
        return result

    def list_projects(self):
        """
        List the available projects.

        Note: To perform this action, user you are currently authenticated with
        needs to be an admin.

        :rtype: ``list`` of :class:`OpenStackIdentityProject`
        """
        response = self.authenticated_request("/v3/projects", method="GET")
        result = self._to_projects(data=response.object["projects"])
        return result

    def list_users(self):
        """
        List the available users.

        :rtype: ``list`` of :class:`.OpenStackIdentityUser`
        """
        response = self.authenticated_request("/v3/users", method="GET")
        result = self._to_users(data=response.object["users"])
        return result

    def list_roles(self):
        """
        List the available roles.

        :rtype: ``list`` of :class:`.OpenStackIdentityRole`
        """
        response = self.authenticated_request("/v3/roles", method="GET")
        result = self._to_roles(data=response.object["roles"])
        return result

    def get_domain(self, domain_id):
        """
        Retrieve information about a single domain.

        :param domain_id: ID of domain to retrieve information for.
        :type domain_id: ``str``

        :rtype: :class:`.OpenStackIdentityDomain`
        """
        response = self.authenticated_request("/v3/domains/%s" % (domain_id), method="GET")
        result = self._to_domain(data=response.object["domain"])
        return result

    def get_user(self, user_id):
        """
        Get a user account by ID.

        :param user_id: User's id.
        :type name: ``str``

        :return: Located user.
        :rtype: :class:`.OpenStackIdentityUser`
        """
        response = self.authenticated_request("/v3/users/%s" % user_id)
        user = self._to_user(data=response.object["user"])
        return user

    def list_user_projects(self, user):
        """
        Retrieve all the projects user belongs to.

        :rtype: ``list`` of :class:`.OpenStackIdentityProject`
        """
        path = "/v3/users/%s/projects" % (user.id)
        response = self.authenticated_request(path, method="GET")
        result = self._to_projects(data=response.object["projects"])
        return result

    def list_user_domain_roles(self, domain, user):
        """
        Retrieve all the roles for a particular user on a domain.

        :rtype: ``list`` of :class:`.OpenStackIdentityRole`
        """
        # TODO: Also add "get users roles" and "get assginements" which are
        # available in 3.1 and 3.3
        path = "/v3/domains/{}/users/{}/roles".format(domain.id, user.id)
        response = self.authenticated_request(path, method="GET")
        result = self._to_roles(data=response.object["roles"])
        return result

    def grant_domain_role_to_user(self, domain, role, user):
        """
        Grant domain role to a user.

        Note: This function appears to be idempotent.

        :param domain: Domain to grant the role to.
        :type domain: :class:`.OpenStackIdentityDomain`

        :param role: Role to grant.
        :type role: :class:`.OpenStackIdentityRole`

        :param user: User to grant the role to.
        :type user: :class:`.OpenStackIdentityUser`

        :return: ``True`` on success.
        :rtype: ``bool``
        """
        path = "/v3/domains/{}/users/{}/roles/{}".format(domain.id, user.id, role.id)
        response = self.authenticated_request(path, method="PUT")
        return response.status == httplib.NO_CONTENT

    def revoke_domain_role_from_user(self, domain, user, role):
        """
        Revoke domain role from a user.

        :param domain: Domain to revoke the role from.
        :type domain: :class:`.OpenStackIdentityDomain`

        :param role: Role to revoke.
        :type role: :class:`.OpenStackIdentityRole`

        :param user: User to revoke the role from.
        :type user: :class:`.OpenStackIdentityUser`

        :return: ``True`` on success.
        :rtype: ``bool``
        """
        path = "/v3/domains/{}/users/{}/roles/{}".format(domain.id, user.id, role.id)
        response = self.authenticated_request(path, method="DELETE")
        return response.status == httplib.NO_CONTENT

    def grant_project_role_to_user(self, project, role, user):
        """
        Grant project role to a user.

        Note: This function appears to be idempotent.

        :param project: Project to grant the role to.
        :type project: :class:`.OpenStackIdentityDomain`

        :param role: Role to grant.
        :type role: :class:`.OpenStackIdentityRole`

        :param user: User to grant the role to.
        :type user: :class:`.OpenStackIdentityUser`

        :return: ``True`` on success.
        :rtype: ``bool``
        """
        path = "/v3/projects/{}/users/{}/roles/{}".format(project.id, user.id, role.id)
        response = self.authenticated_request(path, method="PUT")
        return response.status == httplib.NO_CONTENT

    def revoke_project_role_from_user(self, project, role, user):
        """
        Revoke project role from a user.

        :param project: Project to revoke the role from.
        :type project: :class:`.OpenStackIdentityDomain`

        :param role: Role to revoke.
        :type role: :class:`.OpenStackIdentityRole`

        :param user: User to revoke the role from.
        :type user: :class:`.OpenStackIdentityUser`

        :return: ``True`` on success.
        :rtype: ``bool``
        """
        path = "/v3/projects/{}/users/{}/roles/{}".format(project.id, user.id, role.id)
        response = self.authenticated_request(path, method="DELETE")
        return response.status == httplib.NO_CONTENT

    def create_user(
        self,
        email,
        password,
        name,
        description=None,
        domain_id=None,
        default_project_id=None,
        enabled=True,
    ):
        """
        Create a new user account.

        :param email: User's mail address.
        :type email: ``str``

        :param password: User's password.
        :type password: ``str``

        :param name: User's name.
        :type name: ``str``

        :param description: Optional description.
        :type description: ``str``

        :param domain_id: ID of the domain to add the user to (optional).
        :type domain_id: ``str``

        :param default_project_id: ID of the default user project (optional).
        :type default_project_id: ``str``

        :param enabled: True to enable user after creation.
        :type enabled: ``bool``

        :return: Created user.
        :rtype: :class:`.OpenStackIdentityUser`
        """
        data = {"email": email, "password": password, "name": name, "enabled": enabled}

        if description:
            data["description"] = description

        if domain_id:
            data["domain_id"] = domain_id

        if default_project_id:
            data["default_project_id"] = default_project_id

        data = json.dumps({"user": data})
        response = self.authenticated_request("/v3/users", data=data, method="POST")

        user = self._to_user(data=response.object["user"])
        return user

    def enable_user(self, user):
        """
        Enable user account.

        Note: This operation appears to be idempotent.

        :param user: User to enable.
        :type user: :class:`.OpenStackIdentityUser`

        :return: User account which has been enabled.
        :rtype: :class:`.OpenStackIdentityUser`
        """
        data = {"enabled": True}
        data = json.dumps({"user": data})
        response = self.authenticated_request("/v3/users/%s" % (user.id), data=data, method="PATCH")

        user = self._to_user(data=response.object["user"])
        return user

    def disable_user(self, user):
        """
        Disable user account.

        Note: This operation appears to be idempotent.

        :param user: User to disable.
        :type user: :class:`.OpenStackIdentityUser`

        :return: User account which has been disabled.
        :rtype: :class:`.OpenStackIdentityUser`
        """
        data = {"enabled": False}
        data = json.dumps({"user": data})
        response = self.authenticated_request("/v3/users/%s" % (user.id), data=data, method="PATCH")

        user = self._to_user(data=response.object["user"])
        return user

    def _get_auth_data(self):
        data = {
            "auth": {
                "identity": {
                    "methods": ["password"],
                    "password": {
                        "user": {
                            "domain": {"name": self.domain_name},
                            "name": self.user_id,
                            "password": self.key,
                        }
                    },
                }
            }
        }

        if self.token_scope == OpenStackIdentityTokenScope.PROJECT:
            # Scope token to project (tenant)
            data["auth"]["scope"] = {
                "project": {
                    "domain": {"id": self.tenant_domain_id},
                    "name": self.tenant_name,
                }
            }
        elif self.token_scope == OpenStackIdentityTokenScope.DOMAIN:
            # Scope token to domain
            data["auth"]["scope"] = {"domain": {"name": self.domain_name}}
        elif self.token_scope == OpenStackIdentityTokenScope.UNSCOPED:
            pass
        else:
            raise ValueError("Token needs to be scoped either to project or " "a domain")

        return data

    def _load_auth_context_from_cache(self):
        context = super()._load_auth_context_from_cache()
        if context is None:
            return None

        # Since v3 only caches the token and expiration, fetch the
        # service catalog and other bits of the authentication context
        # from Keystone.
        try:
            self._fetch_auth_token()
        except InvalidCredsError:
            # Unauthorized; cached auth context was cleared as part of
            # _fetch_auth_token
            return None

        # Local auth context variables set in _fetch_auth_token
        return context

    def _parse_token_response(self, response, cache_it=False, raise_ambiguous_version_error=True):
        """
        Parse a response from /v3/auth/tokens.

        :param cache_it: Should we cache the authentication context?
        :type cache_it: ``bool``

        :param raise_ambiguous_version_error: Should an ambiguous version
            error be raised on a 300 response?
        :type raise_ambiguous_version_error: ``bool``
        """
        if response.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError()
        elif response.status in [httplib.OK, httplib.CREATED]:
            headers = response.headers

            try:
                body = json.loads(response.body)
            except Exception as e:
                raise MalformedResponseError("Failed to parse JSON", e)

            try:
                roles = self._to_roles(body["token"]["roles"])
            except Exception:
                roles = []

            try:
                expires = parse_date(body["token"]["expires_at"])
                token = headers["x-subject-token"]

                # Cache the fewest fields required for token reuse to minimize
                # cache size. Other fields, especially the service catalog, can
                # be quite large. Fetch these from Keystone when the token is
                # first loaded from cache.
                if cache_it:
                    self._cache_auth_context(
                        OpenStackAuthenticationContext(token, expiration=expires)
                    )

                self.auth_token = token
                self.auth_token_expires = expires
                # Note: catalog is not returned for unscoped tokens
                self.urls = body["token"].get("catalog", None)
                self.auth_user_info = body["token"].get("user", None)
                self.auth_user_roles = roles
            except KeyError as e:
                raise MalformedResponseError(
                    "Auth JSON response is \
                                             missing required elements",
                    e,
                )
        elif raise_ambiguous_version_error and response.status == 300:
            # ambiguous version request
            raise LibcloudError(
                "Auth request returned ambiguous version error, try"
                "using the version specific URL to connect,"
                " e.g. identity/v3/auth/tokens"
            )
        else:
            body = "code: {} body:{}".format(response.status, response.body)
            raise MalformedResponseError("Malformed response", body=body, driver=self.driver)

    def _fetch_auth_token(self):
        """
        Fetch our authentication token and service catalog.
        """
        headers = {"X-Subject-Token": self.auth_token}
        response = self.authenticated_request("/v3/auth/tokens", headers=headers)
        self._parse_token_response(response)
        return self

    def _to_domains(self, data):
        result = []
        for item in data:
            domain = self._to_domain(data=item)
            result.append(domain)

        return result

    def _to_domain(self, data):
        domain = OpenStackIdentityDomain(id=data["id"], name=data["name"], enabled=data["enabled"])
        return domain

    def _to_users(self, data):
        result = []
        for item in data:
            user = self._to_user(data=item)
            result.append(user)

        return result

    def _to_user(self, data):
        user = OpenStackIdentityUser(
            id=data["id"],
            domain_id=data["domain_id"],
            name=data["name"],
            email=data.get("email"),
            description=data.get("description", None),
            enabled=data.get("enabled"),
        )
        return user

    def _to_roles(self, data):
        result = []
        for item in data:
            user = self._to_role(data=item)
            result.append(user)

        return result

    def _to_role(self, data):
        role = OpenStackIdentityRole(
            id=data["id"],
            name=data["name"],
            description=data.get("description", None),
            enabled=data.get("enabled", True),
        )
        return role


class OpenStackIdentity_3_0_Connection_AppCred(OpenStackIdentity_3_0_Connection):
    """
    Connection class for Keystone API v3.x using Application Credentials.

    'user_id' is the application credential id and 'key' is the application
    credential secret.
    """

    name = "OpenStack Identity API v3.x with Application Credentials"

    def __init__(
        self,
        auth_url,
        user_id,
        key,
        tenant_name=None,
        domain_name=None,
        tenant_domain_id=None,
        token_scope=None,
        timeout=None,
        proxy_url=None,
        parent_conn=None,
        auth_cache=None,
    ):
        """
        Tenant, domain and scope options are ignored as they are contained
        within the app credential itself and can't be changed.
        """
        super().__init__(
            auth_url=auth_url,
            user_id=user_id,
            key=key,
            tenant_name=tenant_name,
            domain_name=domain_name,
            token_scope=OpenStackIdentityTokenScope.UNSCOPED,
            timeout=timeout,
            proxy_url=proxy_url,
            parent_conn=parent_conn,
            auth_cache=auth_cache,
        )

    def _get_auth_data(self):
        data = {
            "auth": {
                "identity": {
                    "methods": ["application_credential"],
                    "application_credential": {"id": self.user_id, "secret": self.key},
                }
            }
        }
        return data


class OpenStackIdentity_3_0_Connection_OIDC_access_token(OpenStackIdentity_3_0_Connection):
    """
    Connection class for Keystone API v3.x. using OpenID Connect tokens

    The OIDC token must be set in the self.key attribute.

    The identity provider name required to get the full path
    must be set in the self.user_id attribute.

    The protocol name required to get the full path
    must be set in the self.tenant_name attribute.

    The self.domain_name attribute can be used either to select the
    domain name in case of domain scoped token or to select the project
    name in case of project scoped token
    """

    responseCls = OpenStackAuthResponse
    name = "OpenStack Identity API v3.x with OIDC support"
    auth_version = "3.0"

    def authenticate(self, force=False):
        """
        Perform authentication.
        """
        if not self._is_authentication_needed(force=force):
            return self

        subject_token = self._get_unscoped_token_from_oidc_token()

        data = {"auth": {"identity": {"methods": ["token"], "token": {"id": subject_token}}}}

        if self.token_scope == OpenStackIdentityTokenScope.PROJECT:
            # Scope token to project (tenant)
            project_id = self._get_project_id(token=subject_token)
            data["auth"]["scope"] = {"project": {"id": project_id}}
        elif self.token_scope == OpenStackIdentityTokenScope.DOMAIN:
            # Scope token to domain
            data["auth"]["scope"] = {"domain": {"name": self.domain_name}}
        elif self.token_scope == OpenStackIdentityTokenScope.UNSCOPED:
            pass
        else:
            raise ValueError("Token needs to be scoped either to project or " "a domain")

        data = json.dumps(data)
        response = self.request(
            "/v3/auth/tokens",
            data=data,
            headers={"Content-Type": "application/json"},
            method="POST",
        )
        self._parse_token_response(response, cache_it=True, raise_ambiguous_version_error=False)
        return self

    def _get_unscoped_token_from_oidc_token(self):
        """
        Get unscoped token from OIDC access token
        """
        path = "/v3/OS-FEDERATION/identity_providers/{}/protocols/{}/auth".format(
            self.user_id,
            self.tenant_name,
        )
        response = self.request(
            path,
            headers={
                "Content-Type": "application/json",
                "Authorization": "Bearer %s" % self.key,
            },
            method="GET",
        )

        if response.status == httplib.UNAUTHORIZED:
            # Invalid credentials
            raise InvalidCredsError()
        elif response.status in [httplib.OK, httplib.CREATED]:
            if "x-subject-token" in response.headers:
                return response.headers["x-subject-token"]
            else:
                raise MalformedResponseError("No x-subject-token returned", driver=self.driver)
        else:
            raise MalformedResponseError(
                "Malformed response", driver=self.driver, body=response.body
            )

    def _get_project_id(self, token):
        """
        Get the first project ID accessible with the specified access token
        """
        # Try new path first (from ver 1.1)
        path = "/v3/auth/projects"
        response = self.request(
            path,
            headers={"Content-Type": "application/json", AUTH_TOKEN_HEADER: token},
            method="GET",
        )

        if response.status not in [httplib.UNAUTHORIZED, httplib.OK, httplib.CREATED]:
            # In case of error try old one
            path = "/v3/OS-FEDERATION/projects"
            response = self.request(
                path,
                headers={"Content-Type": "application/json", AUTH_TOKEN_HEADER: token},
                method="GET",
            )

        if response.status == httplib.UNAUTHORIZED:
            # Invalid credentials
            raise InvalidCredsError()
        elif response.status in [httplib.OK, httplib.CREATED]:
            try:
                body = json.loads(response.body)
                # We use domain_name in both cases of the scoped tokens
                # as we have used tenant as the protocol
                if self.domain_name and self.domain_name != "Default":
                    for project in body["projects"]:
                        if self.domain_name in [project["name"], project["id"]]:
                            return project["id"]
                    raise ValueError("Project %s not found" % (self.domain_name))
                else:
                    return body["projects"][0]["id"]
            except ValueError as e:
                raise e
            except Exception as e:
                raise MalformedResponseError("Failed to parse JSON", e)
        else:
            raise MalformedResponseError(
                "Malformed response", driver=self.driver, body=response.body
            )


class OpenStackIdentity_2_0_Connection_VOMS(OpenStackIdentityConnection, CertificateConnection):
    """
    Connection class for Keystone API v2.0. with VOMS proxy support
    In this case the key parameter will be the path of the VOMS proxy file.
    """

    responseCls = OpenStackAuthResponse
    name = "OpenStack Identity API v2.0 VOMS support"
    auth_version = "2.0"

    def __init__(
        self,
        auth_url,
        user_id,
        key,
        tenant_name=None,
        tenant_domain_id="default",
        domain_name="Default",
        token_scope=OpenStackIdentityTokenScope.PROJECT,
        timeout=None,
        proxy_url=None,
        parent_conn=None,
        auth_cache=None,
    ):
        CertificateConnection.__init__(
            self, cert_file=key, url=auth_url, proxy_url=proxy_url, timeout=timeout
        )

        self.parent_conn = parent_conn

        # enable tests to use the same mock connection classes.
        if parent_conn:
            self.conn_class = parent_conn.conn_class
            self.driver = parent_conn.driver
        else:
            self.driver = None

        self.auth_url = auth_url
        self.tenant_name = tenant_name
        self.tenant_domain_id = tenant_domain_id
        self.domain_name = domain_name
        self.token_scope = token_scope
        self.timeout = timeout
        self.proxy_url = proxy_url
        self.auth_cache = auth_cache

        self.urls = {}
        self.auth_token = None
        self.auth_token_expires = None
        self.auth_user_info = None

    def authenticate(self, force=False):
        if not self._is_authentication_needed(force=force):
            return self

        tenant = self.tenant_name
        if not tenant:
            # if the tenant name is not specified look for it
            token = self._get_unscoped_token()
            tenant = self._get_tenant_name(token)

        data = {"auth": {"voms": True, "tenantName": tenant}}

        reqbody = json.dumps(data)
        return self._authenticate_2_0_with_body(reqbody)

    def _get_unscoped_token(self):
        """
        Get unscoped token from VOMS proxy
        """
        data = {"auth": {"voms": True}}
        reqbody = json.dumps(data)

        response = self.request(
            "/v2.0/tokens",
            data=reqbody,
            headers={"Content-Type": "application/json"},
            method="POST",
        )

        if response.status == httplib.UNAUTHORIZED:
            # Invalid credentials
            raise InvalidCredsError()
        elif response.status in [httplib.OK, httplib.CREATED]:
            try:
                body = json.loads(response.body)
                return body["access"]["token"]["id"]
            except Exception as e:
                raise MalformedResponseError("Failed to parse JSON", e)
        else:
            raise MalformedResponseError(
                "Malformed response", driver=self.driver, body=response.body
            )

    def _get_tenant_name(self, token):
        """
        Get the first available tenant name (usually there are only one)
        """
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            AUTH_TOKEN_HEADER: token,
        }
        response = self.request("/v2.0/tenants", headers=headers, method="GET")

        if response.status == httplib.UNAUTHORIZED:
            # Invalid credentials
            raise InvalidCredsError()
        elif response.status in [httplib.OK, httplib.CREATED]:
            try:
                body = json.loads(response.body)
                return body["tenants"][0]["name"]
            except Exception as e:
                raise MalformedResponseError("Failed to parse JSON", e)
        else:
            raise MalformedResponseError(
                "Malformed response", driver=self.driver, body=response.body
            )

    def _authenticate_2_0_with_body(self, reqbody):
        resp = self.request(
            "/v2.0/tokens",
            data=reqbody,
            headers={"Content-Type": "application/json"},
            method="POST",
        )

        if resp.status == httplib.UNAUTHORIZED:
            raise InvalidCredsError()
        elif resp.status not in [httplib.OK, httplib.NON_AUTHORITATIVE_INFORMATION]:
            body = "code: {} body: {}".format(resp.status, resp.body)
            raise MalformedResponseError("Malformed response", body=body, driver=self.driver)
        else:
            body = resp.object

            try:
                access = body["access"]
                expires = access["token"]["expires"]
                self._cache_auth_context(
                    OpenStackAuthenticationContext(
                        access["token"]["id"],
                        expiration=parse_date(expires),
                        urls=access["serviceCatalog"],
                        user=access.get("user", {}),
                    )
                )
            except KeyError as e:
                raise MalformedResponseError(
                    "Auth JSON response is \
                                             missing required elements",
                    e,
                )
        return self


def get_class_for_auth_version(auth_version):
    """
    Retrieve class for the provided auth version.
    """
    if auth_version == "1.0":
        cls = OpenStackIdentity_1_0_Connection
    elif auth_version == "1.1":
        cls = OpenStackIdentity_1_1_Connection
    elif auth_version == "2.0" or auth_version == "2.0_apikey":
        cls = OpenStackIdentity_2_0_Connection
    elif auth_version == "2.0_password":
        cls = OpenStackIdentity_2_0_Connection
    elif auth_version == "2.0_voms":
        cls = OpenStackIdentity_2_0_Connection_VOMS
    elif auth_version == "3.x_password":
        cls = OpenStackIdentity_3_0_Connection
    elif auth_version == "3.x_appcred":
        cls = OpenStackIdentity_3_0_Connection_AppCred
    elif auth_version == "3.x_oidc_access_token":
        cls = OpenStackIdentity_3_0_Connection_OIDC_access_token
    else:
        raise LibcloudError("Unsupported Auth Version requested: %s" % (auth_version))

    return cls



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import mimetypes
from typing import Generator

from libcloud.utils.py3 import b, next

CHUNK_SIZE = 8096

__all__ = ["read_in_chunks", "exhaust_iterator", "guess_file_mime_type"]


def read_in_chunks(iterator, chunk_size=None, fill_size=False, yield_empty=False):
    """
    Return a generator which yields data in chunks.

    :param iterator: An object which implements an iterator interface
                     or a File like object with read method.
    :type iterator: :class:`object` which implements iterator interface.

    :param chunk_size: Optional chunk size (defaults to CHUNK_SIZE)
    :type chunk_size: ``int``

    :param fill_size: If True, make sure chunks are exactly chunk_size in
                      length (except for last chunk).
    :type fill_size: ``bool``

    :param yield_empty: If true and iterator returned no data, only yield empty
                        bytes object
    :type yield_empty: ``bool``

    TODO: At some point in the future we could use byte arrays here if version
    >= Python 3. This should speed things up a bit and reduce memory usage.
    """
    chunk_size = chunk_size or CHUNK_SIZE

    try:
        get_data = iterator.read
        args = (chunk_size,)
    except AttributeError:
        get_data = next
        args = (iterator,)

    data = b("")
    empty = False

    while not empty or len(data) > 0:
        if not empty:
            try:
                chunk = b(get_data(*args))
                if len(chunk) > 0:
                    data += chunk
                else:
                    empty = True
            except StopIteration:
                empty = True

        if len(data) == 0:
            if empty and yield_empty:
                yield b("")

            return

        if fill_size:
            data = yield from _optimized_chunked_generator(data=data, chunk_size=chunk_size)
            if empty:
                # Yield last not completely filled chunk
                yield data[:chunk_size]
                data = data[chunk_size:]
        else:
            yield data
            data = b("")


def _optimized_chunked_generator(data: bytes, chunk_size: int) -> Generator[bytes, None, bytes]:
    # We want to emit chunk_size large chunks, but chunk_size can be larger or smaller than the chunks returned
    # by get_data. We need to yield in a loop to avoid large amounts of data piling up.
    # The loop also avoids copying all data #chunks amount of times by keeping the original data as is.
    chunk_start = 0
    while chunk_start + chunk_size < len(data):
        yield data[chunk_start : chunk_start + chunk_size]
        chunk_start += chunk_size
    data = data[chunk_start:]
    return data


def exhaust_iterator(iterator):
    """
    Exhaust an iterator and return all data returned by it.

    :type iterator: :class:`object` which implements iterator interface.
    :param iterator: An object which implements an iterator interface
                     or a File like object with read method.

    :rtype ``str``
    :return Data returned by the iterator.
    """
    data = b("")

    try:
        chunk = b(next(iterator))
    except StopIteration:
        chunk = b("")

    while len(chunk) > 0:
        data += chunk

        try:
            chunk = b(next(iterator))
        except StopIteration:
            chunk = b("")

    return data


def guess_file_mime_type(file_path):
    filename = os.path.basename(file_path)
    mimetype, encoding = mimetypes.guess_type(filename)
    return mimetype, encoding



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["ZerigoDNSDriver"]


import copy
import base64

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import Provider, RecordType, ZoneDoesNotExistError, RecordDoesNotExistError
from libcloud.utils.py3 import ET, b, httplib
from libcloud.utils.xml import findall, findtext
from libcloud.utils.misc import get_new_obj, merge_valid_keys
from libcloud.common.base import XmlResponse, ConnectionUserAndKey
from libcloud.common.types import LibcloudError, InvalidCredsError, MalformedResponseError

API_HOST = "ns.zerigo.com"
API_VERSION = "1.1"
API_ROOT = "/api/%s/" % (API_VERSION)

VALID_ZONE_EXTRA_PARAMS = ["notes", "tag-list", "ns1", "slave-nameservers"]
VALID_RECORD_EXTRA_PARAMS = ["notes", "ttl", "priority"]

# Number of items per page (maximum limit is 1000)
ITEMS_PER_PAGE = 100


class ZerigoError(LibcloudError):
    def __init__(self, code, errors):
        self.code = code
        self.errors = errors or []

    def __str__(self):
        return "Errors: %s" % (", ".join(self.errors))

    def __repr__(self):
        return "<ZerigoError response code={}, errors count={}>".format(
            self.code,
            len(self.errors),
        )


class ZerigoDNSResponse(XmlResponse):
    def success(self):
        return self.status in [httplib.OK, httplib.CREATED, httplib.ACCEPTED]

    def parse_error(self):
        status = int(self.status)

        if status == 401:
            if not self.body:
                raise InvalidCredsError(str(self.status) + ": " + self.error)
            else:
                raise InvalidCredsError(self.body)
        elif status == 404:
            context = self.connection.context

            if context["resource"] == "zone":
                raise ZoneDoesNotExistError(value="", driver=self, zone_id=context["id"])
            elif context["resource"] == "record":
                raise RecordDoesNotExistError(value="", driver=self, record_id=context["id"])
        elif status != 503:
            try:
                body = ET.XML(self.body)
            except Exception:
                raise MalformedResponseError("Failed to parse XML", body=self.body)

            errors = []

            for error in findall(element=body, xpath="error"):
                errors.append(error.text)

            raise ZerigoError(code=status, errors=errors)

        return self.body


class ZerigoDNSConnection(ConnectionUserAndKey):
    host = API_HOST
    secure = True
    responseCls = ZerigoDNSResponse

    def add_default_headers(self, headers):
        auth_b64 = base64.b64encode(b("{}:{}".format(self.user_id, self.key)))
        headers["Authorization"] = "Basic %s" % (auth_b64.decode("utf-8"))

        return headers

    def request(self, action, params=None, data="", headers=None, method="GET"):
        if not headers:
            headers = {}

        if not params:
            params = {}

        if method in ("POST", "PUT"):
            headers = {"Content-Type": "application/xml; charset=UTF-8"}

        return super().request(
            action=action, params=params, data=data, method=method, headers=headers
        )


class ZerigoDNSDriver(DNSDriver):
    type = Provider.ZERIGO
    name = "Zerigo DNS"
    website = "http://www.zerigo.com/"
    connectionCls = ZerigoDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.GEO: "GEO",
        RecordType.MX: "MX",
        RecordType.NAPTR: "NAPTR",
        RecordType.NS: "NS",
        RecordType.PTR: "PTR",
        RecordType.REDIRECT: "REDIRECT",
        RecordType.SPF: "SPF",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
        RecordType.URL: "URL",
    }

    def iterate_zones(self):
        return self._get_more("zones")

    def iterate_records(self, zone):
        return self._get_more("records", zone=zone)

    def get_zone(self, zone_id):
        path = API_ROOT + "zones/%s.xml" % (zone_id)
        self.connection.set_context({"resource": "zone", "id": zone_id})
        data = self.connection.request(path).object
        zone = self._to_zone(elem=data)

        return zone

    def get_record(self, zone_id, record_id):
        zone = self.get_zone(zone_id=zone_id)
        self.connection.set_context({"resource": "record", "id": record_id})
        path = API_ROOT + "hosts/%s.xml" % (record_id)
        data = self.connection.request(path).object
        record = self._to_record(elem=data, zone=zone)

        return record

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Create a new zone.

        Provider API docs:
        https://www.zerigo.com/docs/apis/dns/1.1/zones/create

        @inherits: :class:`DNSDriver.create_zone`
        """
        path = API_ROOT + "zones.xml"
        zone_elem = self._to_zone_elem(domain=domain, type=type, ttl=ttl, extra=extra)
        data = self.connection.request(
            action=path, data=ET.tostring(zone_elem), method="POST"
        ).object
        zone = self._to_zone(elem=data)

        return zone

    def update_zone(self, zone, domain=None, type=None, ttl=None, extra=None):
        """
        Update an existing zone.

        Provider API docs:
        https://www.zerigo.com/docs/apis/dns/1.1/zones/update

        @inherits: :class:`DNSDriver.update_zone`
        """

        if domain:
            raise LibcloudError("Domain cannot be changed", driver=self)

        path = API_ROOT + "zones/%s.xml" % (zone.id)
        zone_elem = self._to_zone_elem(domain=domain, type=type, ttl=ttl, extra=extra)
        response = self.connection.request(action=path, data=ET.tostring(zone_elem), method="PUT")
        assert response.status == httplib.OK

        merged = merge_valid_keys(
            params=copy.deepcopy(zone.extra),
            valid_keys=VALID_ZONE_EXTRA_PARAMS,
            extra=extra,
        )
        updated_zone = get_new_obj(
            obj=zone, klass=Zone, attributes={"type": type, "ttl": ttl, "extra": merged}
        )

        return updated_zone

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        Provider API docs:
        https://www.zerigo.com/docs/apis/dns/1.1/hosts/create

        @inherits: :class:`DNSDriver.create_record`
        """
        path = API_ROOT + "zones/%s/hosts.xml" % (zone.id)
        record_elem = self._to_record_elem(name=name, type=type, data=data, extra=extra)
        response = self.connection.request(
            action=path, data=ET.tostring(record_elem), method="POST"
        )
        assert response.status == httplib.CREATED
        record = self._to_record(elem=response.object, zone=zone)

        return record

    def update_record(self, record, name=None, type=None, data=None, extra=None):
        path = API_ROOT + "hosts/%s.xml" % (record.id)
        record_elem = self._to_record_elem(name=name, type=type, data=data, extra=extra)
        response = self.connection.request(action=path, data=ET.tostring(record_elem), method="PUT")
        assert response.status == httplib.OK

        merged = merge_valid_keys(
            params=copy.deepcopy(record.extra),
            valid_keys=VALID_RECORD_EXTRA_PARAMS,
            extra=extra,
        )
        updated_record = get_new_obj(
            obj=record,
            klass=Record,
            attributes={"type": type, "data": data, "extra": merged},
        )

        return updated_record

    def delete_zone(self, zone):
        path = API_ROOT + "zones/%s.xml" % (zone.id)
        self.connection.set_context({"resource": "zone", "id": zone.id})
        response = self.connection.request(action=path, method="DELETE")

        return response.status == httplib.OK

    def delete_record(self, record):
        path = API_ROOT + "hosts/%s.xml" % (record.id)
        self.connection.set_context({"resource": "record", "id": record.id})
        response = self.connection.request(action=path, method="DELETE")

        return response.status == httplib.OK

    def ex_get_zone_by_domain(self, domain):
        """
        Retrieve a zone object by the domain name.

        :param domain: The domain which should be used
        :type  domain: ``str``

        :rtype: :class:`Zone`
        """
        path = API_ROOT + "zones/%s.xml" % (domain)
        self.connection.set_context({"resource": "zone", "id": domain})
        data = self.connection.request(path).object
        zone = self._to_zone(elem=data)

        return zone

    def ex_force_slave_axfr(self, zone):
        """
        Force a zone transfer.

        :param zone: Zone which should be used.
        :type  zone: :class:`Zone`

        :rtype: :class:`Zone`
        """
        path = API_ROOT + "zones/%s/force_slave_axfr.xml" % (zone.id)
        self.connection.set_context({"resource": "zone", "id": zone.id})
        response = self.connection.request(path, method="POST")
        assert response.status == httplib.ACCEPTED

        return zone

    def _to_zone_elem(self, domain=None, type=None, ttl=None, extra=None):
        zone_elem = ET.Element("zone", {})

        if domain:
            domain_elem = ET.SubElement(zone_elem, "domain")
            domain_elem.text = domain

        if type:
            ns_type_elem = ET.SubElement(zone_elem, "ns-type")

            if type == "master":
                ns_type_elem.text = "pri_sec"
            elif type == "slave":
                if not extra or "ns1" not in extra:
                    raise LibcloudError(
                        "ns1 extra attribute is required " + "when zone type is slave",
                        driver=self,
                    )

                ns_type_elem.text = "sec"
                ns1_elem = ET.SubElement(zone_elem, "ns1")
                ns1_elem.text = extra["ns1"]
            elif type == "std_master":
                # TODO: Each driver should provide supported zone types
                # Slave name servers are elsewhere

                if not extra or "slave-nameservers" not in extra:
                    raise LibcloudError(
                        "slave-nameservers extra "
                        + "attribute is required whenzone "
                        + "type is std_master",
                        driver=self,
                    )

                ns_type_elem.text = "pri"
                slave_nameservers_elem = ET.SubElement(zone_elem, "slave-nameservers")
                slave_nameservers_elem.text = extra["slave-nameservers"]

        if ttl:
            default_ttl_elem = ET.SubElement(zone_elem, "default-ttl")
            default_ttl_elem.text = str(ttl)

        if extra and "tag-list" in extra:
            tags = extra["tag-list"]

            tags_elem = ET.SubElement(zone_elem, "tag-list")
            tags_elem.text = " ".join(tags)

        return zone_elem

    def _to_record_elem(self, name=None, type=None, data=None, extra=None):
        record_elem = ET.Element("host", {})

        if name:
            name_elem = ET.SubElement(record_elem, "hostname")
            name_elem.text = name

        if type is not None:
            type_elem = ET.SubElement(record_elem, "host-type")
            type_elem.text = self.RECORD_TYPE_MAP[type]

        if data:
            data_elem = ET.SubElement(record_elem, "data")
            data_elem.text = data

        if extra:
            if "ttl" in extra:
                ttl_elem = ET.SubElement(record_elem, "ttl", {"type": "integer"})
                ttl_elem.text = str(extra["ttl"])

            if "priority" in extra:
                # Only MX and SRV records support priority
                priority_elem = ET.SubElement(record_elem, "priority", {"type": "integer"})

                priority_elem.text = str(extra["priority"])

            if "notes" in extra:
                notes_elem = ET.SubElement(record_elem, "notes")
                notes_elem.text = extra["notes"]

        return record_elem

    def _to_zones(self, elem):
        zones = []

        for item in findall(element=elem, xpath="zone"):
            zone = self._to_zone(elem=item)
            zones.append(zone)

        return zones

    def _to_zone(self, elem):
        id = findtext(element=elem, xpath="id")
        domain = findtext(element=elem, xpath="domain")
        type = findtext(element=elem, xpath="ns-type")
        type = "master" if type.find("pri") == 0 else "slave"
        ttl = findtext(element=elem, xpath="default-ttl")

        hostmaster = findtext(element=elem, xpath="hostmaster")
        custom_ns = findtext(element=elem, xpath="custom-ns")
        custom_nameservers = findtext(element=elem, xpath="custom-nameservers")
        notes = findtext(element=elem, xpath="notes")
        nx_ttl = findtext(element=elem, xpath="nx-ttl")
        slave_nameservers = findtext(element=elem, xpath="slave-nameservers")
        tags = findtext(element=elem, xpath="tag-list")
        tags = tags.split(" ") if tags else []

        extra = {
            "hostmaster": hostmaster,
            "custom-ns": custom_ns,
            "custom-nameservers": custom_nameservers,
            "notes": notes,
            "nx-ttl": nx_ttl,
            "slave-nameservers": slave_nameservers,
            "tags": tags,
        }
        zone = Zone(id=str(id), domain=domain, type=type, ttl=int(ttl), driver=self, extra=extra)

        return zone

    def _to_records(self, elem, zone):
        records = []

        for item in findall(element=elem, xpath="host"):
            record = self._to_record(elem=item, zone=zone)
            records.append(record)

        return records

    def _to_record(self, elem, zone):
        id = findtext(element=elem, xpath="id")
        name = findtext(element=elem, xpath="hostname")
        type = findtext(element=elem, xpath="host-type")
        type = self._string_to_record_type(type)
        data = findtext(element=elem, xpath="data")

        notes = findtext(element=elem, xpath="notes", no_text_value=None)
        state = findtext(element=elem, xpath="state", no_text_value=None)
        fqdn = findtext(element=elem, xpath="fqdn", no_text_value=None)
        priority = findtext(element=elem, xpath="priority", no_text_value=None)
        ttl = findtext(element=elem, xpath="ttl", no_text_value=None)

        if not name:
            name = None

        if ttl:
            ttl = int(ttl)

        extra = {
            "notes": notes,
            "state": state,
            "fqdn": fqdn,
            "priority": priority,
            "ttl": ttl,
        }

        record = Record(
            id=id,
            name=name,
            type=type,
            data=data,
            zone=zone,
            driver=self,
            ttl=ttl,
            extra=extra,
        )

        return record

    def _get_more(self, rtype, **kwargs):
        exhausted = False
        last_key = None

        while not exhausted:
            items, last_key, exhausted = self._get_data(rtype, last_key, **kwargs)

            yield from items

    def _get_data(self, rtype, last_key, **kwargs):
        # Note: last_key in this case really is a "last_page".
        # TODO: Update base driver and change last_key to something more
        # generic - e.g. marker
        params = {}
        params["per_page"] = ITEMS_PER_PAGE
        params["page"] = last_key + 1 if last_key else 1

        if rtype == "zones":
            path = API_ROOT + "zones.xml"
            response = self.connection.request(path)
            transform_func = self._to_zones
        elif rtype == "records":
            zone = kwargs["zone"]
            path = API_ROOT + "zones/%s/hosts.xml" % (zone.id)
            self.connection.set_context({"resource": "zone", "id": zone.id})
            response = self.connection.request(path, params=params)
            transform_func = self._to_records
        else:
            raise ValueError(f"Unsupported rtype: {rtype}")

        exhausted = False
        result_count = int(response.headers.get("x-query-count", 0))

        if (params["page"] * ITEMS_PER_PAGE) >= result_count:
            exhausted = True

        if response.status == httplib.OK:
            items = transform_func(elem=response.object, **kwargs)

            return items, params["page"], exhausted
        else:
            return [], None, True



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
VMware vSphere driver. Uses pyvmomi - https://github.com/vmware/pyvmomi
Code inspired by https://github.com/vmware/pyvmomi-community-samples

Authors: Dimitris Moraitis, Alex Tsiliris, Markos Gogoulos
"""

import ssl
import json
import time
import atexit
import base64
import asyncio
import hashlib
import logging
import warnings
import functools
import itertools

from libcloud.utils.py3 import httplib
from libcloud.common.base import JsonResponse, ConnectionKey
from libcloud.common.types import LibcloudError, ProviderError, InvalidCredsError
from libcloud.compute.base import Node, NodeSize, NodeImage, NodeDriver, NodeLocation
from libcloud.compute.types import Provider, NodeState
from libcloud.utils.networking import is_public_subnet
from libcloud.common.exceptions import BaseHTTPError

try:
    from pyVim import connect
    from pyVmomi import VmomiSupport, vim, vmodl
    from pyVim.task import WaitForTask

    pyvmomi = True
except ImportError:
    pyvmomi = None


logger = logging.getLogger("libcloud.compute.drivers.vsphere")


def recurse_snapshots(snapshot_list):
    ret = []

    for s in snapshot_list:
        ret.append(s)
        ret += recurse_snapshots(getattr(s, "childSnapshotList", []))

    return ret


def format_snapshots(snapshot_list):
    ret = []

    for s in snapshot_list:
        ret.append(
            {
                "id": s.id,
                "name": s.name,
                "description": s.description,
                "created": s.createTime.strftime("%Y-%m-%d %H:%M"),
                "state": s.state,
            }
        )

    return ret


# 6.5 and older, probably won't work on anything earlier than 4.x
class VSphereNodeDriver(NodeDriver):
    name = "VMware vSphere"
    website = "http://www.vmware.com/products/vsphere/"
    type = Provider.VSPHERE

    NODE_STATE_MAP = {
        "poweredOn": NodeState.RUNNING,
        "poweredOff": NodeState.STOPPED,
        "suspended": NodeState.SUSPENDED,
    }

    def __init__(self, host, username, password, port=443, ca_cert=None, verify_ssl=True):
        """Initialize a connection by providing a hostname,
        username and password
        """

        if pyvmomi is None:
            raise ImportError(
                'Missing "pyvmomi" dependency. '
                "You can install it "
                "using pip - pip install pyvmomi"
            )
        self.host = host
        try:
            if ca_cert is None:
                self.connection = connect.SmartConnect(
                    host=host,
                    port=port,
                    user=username,
                    pwd=password,
                )
            else:
                context = ssl.create_default_context(cafile=ca_cert)
                self.connection = connect.SmartConnect(
                    host=host,
                    port=port,
                    user=username,
                    pwd=password,
                    sslContext=context,
                )
            atexit.register(connect.Disconnect, self.connection)
        except Exception as exc:
            error_message = str(exc).lower()

            if "incorrect user name" in error_message:
                raise InvalidCredsError("Check your username and " "password are valid")

            if "connection refused" in error_message or "is not a vim server" in error_message:
                raise LibcloudError(
                    "Check that the host provided is a " "vSphere installation",
                    driver=self,
                )

            if "name or service not known" in error_message:
                raise LibcloudError("Check that the vSphere host is accessible", driver=self)

            if "certificate verify failed" in error_message and not verify_ssl:
                # bypass self signed certificates
                try:
                    context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                    context.verify_mode = ssl.CERT_NONE
                except ImportError:
                    raise ImportError(
                        "To use self signed certificates, "
                        "please upgrade to python 2.7.11 and "
                        "pyvmomi 6.0.0+"
                    )

                self.connection = connect.SmartConnect(
                    host=host,
                    port=port,
                    user=username,
                    pwd=password,
                    sslContext=context,
                )
                atexit.register(connect.Disconnect, self.connection)
            else:
                raise LibcloudError("Cannot connect to vSphere", driver=self)

    def list_locations(self, ex_show_hosts_in_drs=True):
        """
        Lists locations
        """
        content = self.connection.RetrieveContent()

        potential_locations = [
            dc
            for dc in content.viewManager.CreateContainerView(
                content.rootFolder,
                [vim.ClusterComputeResource, vim.HostSystem],
                recursive=True,
            ).view
        ]

        # Add hosts and clusters with DRS enabled
        locations = []
        hosts_all = []
        clusters = []

        for location in potential_locations:
            if isinstance(location, vim.HostSystem):
                hosts_all.append(location)
            elif isinstance(location, vim.ClusterComputeResource):
                if location.configuration.drsConfig.enabled:
                    clusters.append(location)

        if ex_show_hosts_in_drs:
            hosts = hosts_all
        else:
            hosts_filter = [host for cluster in clusters for host in cluster.host]
            hosts = [host for host in hosts_all if host not in hosts_filter]

        for cluster in clusters:
            locations.append(self._to_location(cluster))

        for host in hosts:
            locations.append(self._to_location(host))

        return locations

    def _to_location(self, data):
        extra = {}
        try:
            if isinstance(data, vim.HostSystem):
                extra = {
                    "type": "host",
                    "state": data.runtime.connectionState,
                    "hypervisor": data.config.product.fullName,
                    "vendor": data.hardware.systemInfo.vendor,
                    "model": data.hardware.systemInfo.model,
                    "ram": data.hardware.memorySize,
                    "cpu": {
                        "packages": data.hardware.cpuInfo.numCpuPackages,
                        "cores": data.hardware.cpuInfo.numCpuCores,
                        "threads": data.hardware.cpuInfo.numCpuThreads,
                    },
                    "uptime": data.summary.quickStats.uptime,
                    "parent": str(data.parent),
                }
            elif isinstance(data, vim.ClusterComputeResource):
                extra = {
                    "type": "cluster",
                    "overallStatus": data.overallStatus,
                    "drs": data.configuration.drsConfig.enabled,
                    "hosts": [host.name for host in data.host],
                    "parent": str(data.parent),
                }
        except AttributeError as exc:
            logger.error("Cannot convert location {}: {!r}".format(data.name, exc))
            extra = {}

        return NodeLocation(id=data.name, name=data.name, country=None, extra=extra, driver=self)

    def ex_list_networks(self):
        """
        List networks
        """
        content = self.connection.RetrieveContent()
        networks = content.viewManager.CreateContainerView(
            content.rootFolder, [vim.Network], recursive=True
        ).view

        return [self._to_network(network) for network in networks]

    def _to_network(self, data):
        summary = data.summary
        extra = {
            "hosts": [h.name for h in data.host],
            "ip_pool_name": summary.ipPoolName,
            "ip_pool_id": summary.ipPoolId,
            "accessible": summary.accessible,
        }

        return VSphereNetwork(id=data.name, name=data.name, extra=extra)

    def list_sizes(self):
        """
        Returns sizes
        """

        return []

    def list_images(self, location=None, folder_ids=None):
        """
        Lists VM templates as images.
        If folder is given then it will list images contained
        in that folder only.
        """

        images = []

        if folder_ids:
            vms = []

            for folder_id in folder_ids:
                folder_object = self._get_item_by_moid("Folder", folder_id)
                vms.extend(folder_object.childEntity)
        else:
            content = self.connection.RetrieveContent()
            vms = content.viewManager.CreateContainerView(
                content.rootFolder, [vim.VirtualMachine], recursive=True
            ).view

        for vm in vms:
            if vm.config and vm.config.template:
                images.append(self._to_image(vm))

        return images

    def _to_image(self, data):
        summary = data.summary
        name = summary.config.name
        uuid = summary.config.instanceUuid
        memory = summary.config.memorySizeMB
        cpus = summary.config.numCpu
        operating_system = summary.config.guestFullName
        os_type = "unix"

        if "Microsoft" in str(operating_system):
            os_type = "windows"
        annotation = summary.config.annotation
        extra = {
            "path": summary.config.vmPathName,
            "operating_system": operating_system,
            "os_type": os_type,
            "memory_MB": memory,
            "cpus": cpus,
            "overallStatus": str(summary.overallStatus),
            "metadata": {},
            "type": "template_6_5",
            "disk_size": int(summary.storage.committed) // (1024**3),
            "datastore": data.datastore[0].info.name,
        }

        boot_time = summary.runtime.bootTime

        if boot_time:
            extra["boot_time"] = boot_time.isoformat()

        if annotation:
            extra["annotation"] = annotation

        for custom_field in data.customValue:
            key_id = custom_field.key
            key = self.find_custom_field_key(key_id)
            extra["metadata"][key] = custom_field.value

        return NodeImage(id=uuid, name=name, driver=self, extra=extra)

    def _collect_properties(self, content, view_ref, obj_type, path_set=None, include_mors=False):
        """
        Collect properties for managed objects from a view ref
        Check the vSphere API documentation for example on retrieving
        object properties:
            - http://goo.gl/erbFDz
        Args:
            content     (ServiceInstance): ServiceInstance content
            view_ref (pyVmomi.vim.view.*): Starting point of inventory
                                           navigation
            obj_type      (pyVmomi.vim.*): Type of managed object
            path_set               (list): List of properties to retrieve
            include_mors           (bool): If True include the managed objects
                                        refs in the result
        Returns:
            A list of properties for the managed objects
        """
        collector = content.propertyCollector

        # Create object specification to define the starting point of
        # inventory navigation
        obj_spec = vmodl.query.PropertyCollector.ObjectSpec()
        obj_spec.obj = view_ref
        obj_spec.skip = True

        # Create a traversal specification to identify the path for collection
        traversal_spec = vmodl.query.PropertyCollector.TraversalSpec()
        traversal_spec.name = "traverseEntities"
        traversal_spec.path = "view"
        traversal_spec.skip = False
        traversal_spec.type = view_ref.__class__
        obj_spec.selectSet = [traversal_spec]

        # Identify the properties to the retrieved
        property_spec = vmodl.query.PropertyCollector.PropertySpec()
        property_spec.type = obj_type

        if not path_set:
            property_spec.all = True

        property_spec.pathSet = path_set

        # Add the object and property specification to the
        # property filter specification
        filter_spec = vmodl.query.PropertyCollector.FilterSpec()
        filter_spec.objectSet = [obj_spec]
        filter_spec.propSet = [property_spec]

        # Retrieve properties
        props = collector.RetrieveContents([filter_spec])

        data = []

        for obj in props:
            properties = {}

            for prop in obj.propSet:
                properties[prop.name] = prop.val

            if include_mors:
                properties["obj"] = obj.obj

            data.append(properties)

        return data

    def list_nodes(self, enhance=True, max_properties=20):
        """
        List nodes, excluding templates
        """

        vm_properties = [
            "config.template",
            "summary.config.name",
            "summary.config.vmPathName",
            "summary.config.memorySizeMB",
            "summary.config.numCpu",
            "summary.storage.committed",
            "summary.config.guestFullName",
            "summary.runtime.host",
            "summary.config.instanceUuid",
            "summary.config.annotation",
            "summary.runtime.powerState",
            "summary.runtime.bootTime",
            "summary.guest.ipAddress",
            "summary.overallStatus",
            "customValue",
            "snapshot",
        ]
        content = self.connection.RetrieveContent()
        view = content.viewManager.CreateContainerView(
            content.rootFolder, [vim.VirtualMachine], True
        )
        i = 0
        vm_dict = {}

        while i < len(vm_properties):
            vm_list = self._collect_properties(
                content,
                view,
                vim.VirtualMachine,
                path_set=vm_properties[i : i + max_properties],
                include_mors=True,
            )
            i += max_properties

            for vm in vm_list:
                if not vm_dict.get(vm["obj"]):
                    vm_dict[vm["obj"]] = vm
                else:
                    vm_dict[vm["obj"]].update(vm)

        vm_list = [vm_dict[k] for k in vm_dict]
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        nodes = loop.run_until_complete(self._to_nodes(vm_list))

        if enhance:
            nodes = self._enhance_metadata(nodes, content)

        return nodes

    def list_nodes_recursive(self, enhance=True):
        """
        Lists nodes, excluding templates
        """
        nodes = []
        content = self.connection.RetrieveContent()
        children = content.rootFolder.childEntity
        # this will be needed for custom VM metadata

        if content.customFieldsManager:
            self.custom_fields = content.customFieldsManager.field
        else:
            self.custom_fields = []

        for child in children:
            if hasattr(child, "vmFolder"):
                datacenter = child
                vm_folder = datacenter.vmFolder
                vm_list = vm_folder.childEntity
                nodes.extend(self._to_nodes_recursive(vm_list))

        if enhance:
            nodes = self._enhance_metadata(nodes, content)

        return nodes

    def _enhance_metadata(self, nodes, content):
        nodemap = {}

        for node in nodes:
            node.extra["vSphere version"] = content.about.version
            nodemap[node.id] = node

        # Get VM deployment events to extract creation dates & images
        filter_spec = vim.event.EventFilterSpec(eventTypeId=["VmBeingDeployedEvent"])
        deploy_events = content.eventManager.QueryEvent(filter_spec)

        for event in deploy_events:
            try:
                uuid = event.vm.vm.config.instanceUuid
            except Exception:
                continue

            if uuid in nodemap:
                node = nodemap[uuid]
                try:  # Get source template as image
                    source_template_vm = event.srcTemplate.vm
                    image_id = source_template_vm.config.instanceUuid
                    node.extra["image_id"] = image_id
                except Exception:
                    logger.error("Cannot get instanceUuid " "from source template")
                try:  # Get creation date
                    node.created_at = event.createdTime
                except AttributeError:
                    logger.error("Cannot get creation date from VM " "deploy event")

        return nodes

    async def _to_nodes(self, vm_list):
        vms = []

        for vm in vm_list:
            if vm.get("config.template"):
                continue  # Do not include templates in node list
            vms.append(vm)
        loop = asyncio.get_event_loop()
        vms = [loop.run_in_executor(None, self._to_node, vms[i]) for i in range(len(vms))]

        return await asyncio.gather(*vms)

    def _to_nodes_recursive(self, vm_list):
        nodes = []

        for virtual_machine in vm_list:
            if hasattr(virtual_machine, "childEntity"):
                # If this is a group it will have children.
                # If it does, recurse into them and then return
                nodes.extend(self._to_nodes_recursive(virtual_machine.childEntity))
            elif isinstance(virtual_machine, vim.VirtualApp):
                # If this is a vApp, it likely contains child VMs
                # (vApps can nest vApps, but it is hardly
                # a common usecase, so ignore that)
                nodes.extend(self._to_nodes_recursive(virtual_machine.vm))
            else:
                if not hasattr(virtual_machine, "config") or (
                    virtual_machine.config and virtual_machine.config.template
                ):
                    continue  # Do not include templates in node list
                nodes.append(self._to_node_recursive(virtual_machine))

        return nodes

    def _to_node(self, vm):
        name = vm.get("summary.config.name")
        path = vm.get("summary.config.vmPathName")
        memory = vm.get("summary.config.memorySizeMB")
        cpus = vm.get("summary.config.numCpu")
        disk = vm.get("summary.storage.committed", 0) // (1024**3)
        id_to_hash = str(memory) + str(cpus) + str(disk)
        size_id = hashlib.md5(id_to_hash.encode("utf-8")).hexdigest()  # nosec
        size_name = name + "-size"
        size_extra = {"cpus": cpus}
        driver = self
        size = NodeSize(
            id=size_id,
            name=size_name,
            ram=memory,
            disk=disk,
            bandwidth=0,
            price=0,
            driver=driver,
            extra=size_extra,
        )
        operating_system = vm.get("summary.config.guestFullName")
        host = vm.get("summary.runtime.host")

        os_type = "unix"

        if "Microsoft" in str(operating_system):
            os_type = "windows"
        uuid = vm.get("summary.config.instanceUuid") or (
            vm.get("obj").config and vm.get("obj").config.instanceUuid
        )

        if not uuid:
            logger.error("No uuid for vm: {}".format(vm))
        annotation = vm.get("summary.config.annotation")
        state = vm.get("summary.runtime.powerState")
        status = self.NODE_STATE_MAP.get(state, NodeState.UNKNOWN)
        boot_time = vm.get("summary.runtime.bootTime")

        ip_addresses = []

        if vm.get("summary.guest.ipAddress"):
            ip_addresses.append(vm.get("summary.guest.ipAddress"))

        overall_status = str(vm.get("summary.overallStatus"))
        public_ips = []
        private_ips = []

        extra = {
            "path": path,
            "operating_system": operating_system,
            "os_type": os_type,
            "memory_MB": memory,
            "cpus": cpus,
            "overall_status": overall_status,
            "metadata": {},
            "snapshots": [],
        }

        if disk:
            extra["disk"] = disk

        if host:
            extra["host"] = host.name
            parent = host.parent

            while parent:
                if isinstance(parent, vim.ClusterComputeResource):
                    extra["cluster"] = parent.name

                    break
                parent = parent.parent

        if boot_time:
            extra["boot_time"] = boot_time.isoformat()

        if annotation:
            extra["annotation"] = annotation

        for ip_address in ip_addresses:
            try:
                if is_public_subnet(ip_address):
                    public_ips.append(ip_address)
                else:
                    private_ips.append(ip_address)
            except Exception:
                # IPV6 not supported
                pass

        if vm.get("snapshot"):
            extra["snapshots"] = format_snapshots(
                recurse_snapshots(vm.get("snapshot").rootSnapshotList)
            )

        for custom_field in vm.get("customValue", []):
            key_id = custom_field.key
            key = self.find_custom_field_key(key_id)
            extra["metadata"][key] = custom_field.value

        node = Node(
            id=uuid,
            name=name,
            state=status,
            size=size,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self,
            extra=extra,
        )
        node._uuid = uuid

        return node

    def _to_node_recursive(self, virtual_machine):
        summary = virtual_machine.summary
        name = summary.config.name
        path = summary.config.vmPathName
        memory = summary.config.memorySizeMB
        cpus = summary.config.numCpu
        disk = ""

        if summary.storage.committed:
            disk = summary.storage.committed / (1024**3)
        id_to_hash = str(memory) + str(cpus) + str(disk)
        size_id = hashlib.md5(id_to_hash.encode("utf-8")).hexdigest()  # nosec
        size_name = name + "-size"
        size_extra = {"cpus": cpus}
        driver = self
        size = NodeSize(
            id=size_id,
            name=size_name,
            ram=memory,
            disk=disk,
            bandwidth=0,
            price=0,
            driver=driver,
            extra=size_extra,
        )
        operating_system = summary.config.guestFullName
        host = summary.runtime.host

        # mist.io needs this metadata
        os_type = "unix"

        if "Microsoft" in str(operating_system):
            os_type = "windows"
        uuid = summary.config.instanceUuid
        annotation = summary.config.annotation
        state = summary.runtime.powerState
        status = self.NODE_STATE_MAP.get(state, NodeState.UNKNOWN)
        boot_time = summary.runtime.bootTime
        ip_addresses = []

        if summary.guest is not None:
            ip_addresses.append(summary.guest.ipAddress)

        overall_status = str(summary.overallStatus)
        public_ips = []
        private_ips = []

        extra = {
            "path": path,
            "operating_system": operating_system,
            "os_type": os_type,
            "memory_MB": memory,
            "cpus": cpus,
            "overallStatus": overall_status,
            "metadata": {},
            "snapshots": [],
        }

        if disk:
            extra["disk"] = disk

        if host:
            extra["host"] = host.name
            parent = host.parent

            while parent:
                if isinstance(parent, vim.ClusterComputeResource):
                    extra["cluster"] = parent.name

                    break
                parent = parent.parent

        if boot_time:
            extra["boot_time"] = boot_time.isoformat()

        if annotation:
            extra["annotation"] = annotation

        for ip_address in ip_addresses:
            try:
                if is_public_subnet(ip_address):
                    public_ips.append(ip_address)
                else:
                    private_ips.append(ip_address)
            except Exception:
                # IPV6 not supported
                pass

        if virtual_machine.snapshot:
            snapshots = [
                {
                    "id": s.id,
                    "name": s.name,
                    "description": s.description,
                    "created": s.createTime.strftime("%Y-%m-%d %H:%M"),
                    "state": s.state,
                }
                for s in virtual_machine.snapshot.rootSnapshotList
            ]
            extra["snapshots"] = snapshots

        for custom_field in virtual_machine.customValue:
            key_id = custom_field.key
            key = self.find_custom_field_key(key_id)
            extra["metadata"][key] = custom_field.value

        node = Node(
            id=uuid,
            name=name,
            state=status,
            size=size,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=self,
            extra=extra,
        )
        node._uuid = uuid

        return node

    def reboot_node(self, node):
        """ """
        vm = self.find_by_uuid(node.id)

        return self.wait_for_task(vm.RebootGuest())

    def destroy_node(self, node):
        """ """
        vm = self.find_by_uuid(node.id)

        if node.state != NodeState.STOPPED:
            self.stop_node(node)

        return self.wait_for_task(vm.Destroy())

    def stop_node(self, node):
        """ """
        vm = self.find_by_uuid(node.id)

        return self.wait_for_task(vm.PowerOff())

    def start_node(self, node):
        """ """
        vm = self.find_by_uuid(node.id)

        return self.wait_for_task(vm.PowerOn())

    def ex_list_snapshots(self, node):
        """
        List node snapshots
        """
        vm = self.find_by_uuid(node.id)

        if not vm.snapshot:
            return []

        return format_snapshots(recurse_snapshots(vm.snapshot.rootSnapshotList))

    def ex_create_snapshot(
        self, node, snapshot_name, description="", dump_memory=False, quiesce=False
    ):
        """
        Create node snapshot
        """
        vm = self.find_by_uuid(node.id)

        return WaitForTask(vm.CreateSnapshot(snapshot_name, description, dump_memory, quiesce))

    def ex_remove_snapshot(self, node, snapshot_name=None, remove_children=True):
        """
        Remove a snapshot from node.
        If snapshot_name is not defined remove the last one.
        """
        vm = self.find_by_uuid(node.id)

        if not vm.snapshot:
            raise LibcloudError(
                "Remove snapshot failed. No snapshots for node %s" % node.name,
                driver=self,
            )
        snapshots = recurse_snapshots(vm.snapshot.rootSnapshotList)

        if not snapshot_name:
            snapshot = snapshots[-1].snapshot
        else:
            for s in snapshots:
                if snapshot_name == s.name:
                    snapshot = s.snapshot

                    break
            else:
                raise LibcloudError("Snapshot `%s` not found" % snapshot_name, driver=self)

        return self.wait_for_task(snapshot.RemoveSnapshot_Task(removeChildren=remove_children))

    def ex_revert_to_snapshot(self, node, snapshot_name=None):
        """
        Revert node to a specific snapshot.
        If snapshot_name is not defined revert to the last one.
        """
        vm = self.find_by_uuid(node.id)

        if not vm.snapshot:
            raise LibcloudError(
                "Revert failed. No snapshots " "for node %s" % node.name, driver=self
            )
        snapshots = recurse_snapshots(vm.snapshot.rootSnapshotList)

        if not snapshot_name:
            snapshot = snapshots[-1].snapshot
        else:
            for s in snapshots:
                if snapshot_name == s.name:
                    snapshot = s.snapshot

                    break
            else:
                raise LibcloudError("Snapshot `%s` not found" % snapshot_name, driver=self)

        return self.wait_for_task(snapshot.RevertToSnapshot_Task())

    def _find_template_by_uuid(self, template_uuid):
        # on version 5.5 and earlier search index won't return a VM
        try:
            template = self.find_by_uuid(template_uuid)
        except LibcloudError:
            content = self.connection.RetrieveContent()
            vms = content.viewManager.CreateContainerView(
                content.rootFolder, [vim.VirtualMachine], recursive=True
            ).view

            for vm in vms:
                if vm.config.instanceUuid == template_uuid:
                    template = vm
        except Exception as exc:
            raise LibcloudError("Error while searching for template: %s" % exc, driver=self)

        if not template:
            raise LibcloudError("Unable to locate VirtualMachine.", driver=self)

        return template

    def find_by_uuid(self, node_uuid):
        """Searches VMs for a given uuid
        returns pyVmomi.VmomiSupport.vim.VirtualMachine
        """
        vm = self.connection.content.searchIndex.FindByUuid(None, node_uuid, True, True)

        if not vm:
            # perhaps it is a moid
            vm = self._get_item_by_moid("VirtualMachine", node_uuid)

            if not vm:
                raise LibcloudError("Unable to locate VirtualMachine.", driver=self)

        return vm

    def find_custom_field_key(self, key_id):
        """Return custom field key name, provided it's id"""

        if not hasattr(self, "custom_fields"):
            content = self.connection.RetrieveContent()

            if content.customFieldsManager:
                self.custom_fields = content.customFieldsManager.field
            else:
                self.custom_fields = []

        for k in self.custom_fields:
            if k.key == key_id:
                return k.name

        return None

    def get_obj(self, vimtype, name):
        """
        Return an object by name, if name is None the
        first found object is returned
        """
        obj = None
        content = self.connection.RetrieveContent()
        container = content.viewManager.CreateContainerView(content.rootFolder, vimtype, True)

        for c in container.view:
            if name:
                if c.name == name:
                    obj = c

                    break
            else:
                obj = c

                break

        return obj

    def wait_for_task(self, task, timeout=1800, interval=10):
        """wait for a vCenter task to finish"""
        start_time = time.time()
        task_done = False

        while not task_done:
            if time.time() - start_time >= timeout:
                raise LibcloudError(
                    "Timeout while waiting " "for import task Id %s" % task.info.id,
                    driver=self,
                )

            if task.info.state == "success":
                if task.info.result and str(task.info.result) != "success":
                    return task.info.result

                return True

            if task.info.state == "error":
                raise LibcloudError(task.info.error.msg, driver=self)
            time.sleep(interval)

    def create_node(
        self,
        name,
        image,
        size,
        location=None,
        ex_cluster=None,
        ex_network=None,
        ex_datacenter=None,
        ex_folder=None,
        ex_resource_pool=None,
        ex_datastore_cluster=None,
        ex_datastore=None,
    ):
        """
        Creates and returns node.

        :keyword    ex_network: Name of a "Network" to connect the VM to ",
        :type       ex_network: ``str``

        """
        template = self._find_template_by_uuid(image.id)

        if ex_cluster:
            cluster_name = ex_cluster
        else:
            cluster_name = location.name
        cluster = self.get_obj([vim.ClusterComputeResource], cluster_name)

        if not cluster:  # It is a host go with it
            cluster = self.get_obj([vim.HostSystem], cluster_name)

        datacenter = None

        if not ex_datacenter:  # Get datacenter from cluster
            parent = cluster.parent

            while parent:
                if isinstance(parent, vim.Datacenter):
                    datacenter = parent

                    break
                parent = parent.parent

        if ex_datacenter or datacenter is None:
            datacenter = self.get_obj([vim.Datacenter], ex_datacenter)

        if ex_folder:
            folder = self.get_obj([vim.Folder], ex_folder)

            if folder is None:
                folder = self._get_item_by_moid("Folder", ex_folder)
        else:
            folder = datacenter.vmFolder

        if ex_resource_pool:
            resource_pool = self.get_obj([vim.ResourcePool], ex_resource_pool)
        else:
            try:
                resource_pool = cluster.resourcePool
            except AttributeError:
                resource_pool = cluster.parent.resourcePool
        devices = []
        vmconf = vim.vm.ConfigSpec(
            numCPUs=int(size.extra.get("cpu", 1)),
            memoryMB=int(size.ram),
            deviceChange=devices,
        )
        datastore = None
        pod = None
        podsel = vim.storageDrs.PodSelectionSpec()

        if ex_datastore_cluster:
            pod = self.get_obj([vim.StoragePod], ex_datastore_cluster)
        else:
            content = self.connection.RetrieveContent()
            pods = content.viewManager.CreateContainerView(
                content.rootFolder, [vim.StoragePod], True
            ).view

            for pod in pods:
                if cluster.name.lower() in pod.name:
                    break
        podsel.storagePod = pod
        storagespec = vim.storageDrs.StoragePlacementSpec()
        storagespec.podSelectionSpec = podsel
        storagespec.type = "create"
        storagespec.folder = folder
        storagespec.resourcePool = resource_pool
        storagespec.configSpec = vmconf

        try:
            content = self.connection.RetrieveContent()
            rec = content.storageResourceManager.RecommendDatastores(storageSpec=storagespec)
            rec_action = rec.recommendations[0].action[0]
            real_datastore_name = rec_action.destination.name
        except Exception:
            real_datastore_name = template.datastore[0].info.name

        datastore = self.get_obj([vim.Datastore], real_datastore_name)

        if ex_datastore:
            datastore = self.get_obj([vim.Datastore], ex_datastore)

            if datastore is None:
                datastore = self._get_item_by_moid("Datastore", ex_datastore)
        elif not datastore:
            datastore = self.get_obj([vim.Datastore], template.datastore[0].info.name)
        add_network = True

        if ex_network and len(template.network) > 0:
            for nets in template.network:
                if template in nets.vm:
                    add_network = False

        if ex_network and add_network:
            nicspec = vim.vm.device.VirtualDeviceSpec()
            nicspec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
            nicspec.device = vim.vm.device.VirtualVmxnet3()
            nicspec.device.wakeOnLanEnabled = True
            nicspec.device.deviceInfo = vim.Description()

            portgroup = self.get_obj([vim.dvs.DistributedVirtualPortgroup], ex_network)

            if portgroup:
                dvs_port_connection = vim.dvs.PortConnection()
                dvs_port_connection.portgroupKey = portgroup.key
                dvs_port_connection.switchUuid = portgroup.config.distributedVirtualSwitch.uuid
                nicspec.device.backing = (
                    vim.vm.device.VirtualEthernetCard.DistributedVirtualPortBackingInfo()
                )
                nicspec.device.backing.port = dvs_port_connection
            else:
                nicspec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()
                nicspec.device.backing.network = self.get_obj([vim.Network], ex_network)
                nicspec.device.backing.deviceName = ex_network
            nicspec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
            nicspec.device.connectable.startConnected = True
            nicspec.device.connectable.connected = True
            nicspec.device.connectable.allowGuestControl = True
            devices.append(nicspec)

        # new_disk_kb = int(size.disk) * 1024 * 1024
        # disk_spec = vim.vm.device.VirtualDeviceSpec()
        # disk_spec.fileOperation = "create"
        # disk_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
        # disk_spec.device = vim.vm.device.VirtualDisk()
        # disk_spec.device.backing = \
        #     vim.vm.device.VirtualDisk.FlatVer2BackingInfo()
        # if size.extra.get('disk_type') == 'thin':
        #     disk_spec.device.backing.thinProvisioned = True
        # disk_spec.device.backing.diskMode = 'persistent'
        # disk_spec.device.capacityInKB = new_disk_kb
        # disk_spec.device.controllerKey = controller.key
        # devices.append(disk_spec)

        clonespec = vim.vm.CloneSpec(config=vmconf)
        relospec = vim.vm.RelocateSpec()
        relospec.datastore = datastore
        relospec.pool = resource_pool

        if location:
            host = self.get_obj([vim.HostSystem], location.name)

            if host:
                relospec.host = host
        clonespec.location = relospec
        clonespec.powerOn = True
        task = template.Clone(folder=folder, name=name, spec=clonespec)

        return self._to_node_recursive(self.wait_for_task(task))

    def ex_connect_network(self, vm, network_name):
        spec = vim.vm.ConfigSpec()

        # add Switch here
        dev_changes = []
        network_spec = vim.vm.device.VirtualDeviceSpec()
        # network_spec.fileOperation = "create"
        network_spec.operation = vim.vm.device.VirtualDeviceSpec.Operation.add
        network_spec.device = vim.vm.device.VirtualVmxnet3()

        network_spec.device.backing = vim.vm.device.VirtualEthernetCard.NetworkBackingInfo()

        network_spec.device.backing.useAutoDetect = False
        network_spec.device.backing.network = self.get_obj([vim.Network], network_name)

        network_spec.device.connectable = vim.vm.device.VirtualDevice.ConnectInfo()
        network_spec.device.connectable.startConnected = True
        network_spec.device.connectable.connected = True
        network_spec.device.connectable.allowGuestControl = True

        dev_changes.append(network_spec)

        spec.deviceChange = dev_changes
        output = vm.ReconfigVM_Task(spec=spec)
        print(output.info)

    def _get_item_by_moid(self, type_, moid):
        vm_obj = VmomiSupport.templateOf(type_)(moid, self.connection._stub)

        return vm_obj

    def ex_list_folders(self):
        content = self.connection.RetrieveContent()
        folders_raw = content.viewManager.CreateContainerView(
            content.rootFolder, [vim.Folder], True
        ).view
        folders = []

        for folder in folders_raw:
            to_add = {"type": list(folder.childType)}
            to_add["name"] = folder.name
            to_add["id"] = folder._moId
            folders.append(to_add)

        return folders

    def ex_list_datastores(self):
        content = self.connection.RetrieveContent()
        datastores_raw = content.viewManager.CreateContainerView(
            content.rootFolder, [vim.Datastore], True
        ).view
        datastores = []

        for dstore in datastores_raw:
            to_add = {"type": dstore.summary.type}
            to_add["name"] = dstore.name
            to_add["id"] = dstore._moId
            to_add["free_space"] = int(dstore.summary.freeSpace)
            to_add["capacity"] = int(dstore.summary.capacity)
            datastores.append(to_add)

        return datastores

    def ex_open_console(self, vm_uuid):
        vm = self.find_by_uuid(vm_uuid)
        ticket = vm.AcquireTicket(ticketType="webmks")

        return "wss://{}:{}/ticket/{}".format(ticket.host, ticket.port, ticket.ticket)

    def _get_version(self):
        content = self.connection.RetrieveContent()

        return content.about.version


class VSphereNetwork:
    """
    Represents information about a VPC (Virtual Private Cloud) network

    Note: This class is VSphere specific.
    """

    def __init__(self, id, name, extra=None):
        self.id = id
        self.name = name
        self.extra = extra or {}

    def __repr__(self):
        return ("<VSphereNetwork: id=%s, name=%s") % (self.id, self.name)


# 6.7
class VSphereResponse(JsonResponse):
    def parse_error(self):
        if self.body:
            message = self.body
            message += "-- code: {}".format(self.status)

            return message

        return self.body


class VSphereConnection(ConnectionKey):
    responseCls = VSphereResponse
    session_token = None

    def add_default_headers(self, headers):
        """
        VSphere needs an initial connection to a specific API endpoint to
        generate a session-token, which will be used for the purpose of
        authenticating for the rest of the session.
        """
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"

        if self.session_token is None:
            to_encode = "{}:{}".format(self.key, self.secret)
            b64_user_pass = base64.b64encode(to_encode.encode())
            headers["Authorization"] = "Basic {}".format(b64_user_pass.decode())
        else:
            headers["vmware-api-session-id"] = self.session_token

        return headers


class VSphereException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message
        self.args = (code, message)

    def __str__(self):
        return "{} {}".format(self.code, self.message)

    def __repr__(self):
        return "VSphereException {} {}".format(self.code, self.message)


class VSphere_REST_NodeDriver(NodeDriver):
    name = "VMware vSphere"
    website = "http://www.vmware.com/products/vsphere/"
    type = Provider.VSPHERE
    connectionCls = VSphereConnection
    session_token = None

    NODE_STATE_MAP = {
        "powered_on": NodeState.RUNNING,
        "powered_off": NodeState.STOPPED,
        "suspended": NodeState.SUSPENDED,
    }

    VALID_RESPONSE_CODES = [
        httplib.OK,
        httplib.ACCEPTED,
        httplib.CREATED,
        httplib.NO_CONTENT,
    ]

    def __init__(self, key, secret=None, secure=True, host=None, port=443, ca_cert=None):
        if not key or not secret:
            raise InvalidCredsError("Please provide both username " "(key) and password (secret).")
        super().__init__(key=key, secure=secure, host=host, port=port)
        prefixes = ["http://", "https://"]

        for prefix in prefixes:
            if host.startswith(prefix):
                host = host.lstrip(prefix)

        if ca_cert:
            self.connection.connection.ca_cert = ca_cert
        else:
            self.connection.connection.ca_cert = False

        self.connection.secret = secret
        self.host = host
        self.username = key
        # getting session token
        self._get_session_token()
        self.driver_soap = None

    def _get_soap_driver(self):
        if pyvmomi is None:
            raise ImportError(
                'Missing "pyvmomi" dependency. '
                "You can install it "
                "using pip - pip install pyvmomi"
            )
        self.driver_soap = VSphereNodeDriver(
            self.host,
            self.username,
            self.connection.secret,
            ca_cert=self.connection.connection.ca_cert,
        )

    def _get_session_token(self):
        uri = "/rest/com/vmware/cis/session"
        try:
            result = self.connection.request(uri, method="POST")
        except Exception:
            raise
        self.session_token = result.object["value"]
        self.connection.session_token = self.session_token

    def list_sizes(self):
        return []

    def list_nodes(
        self,
        ex_filter_power_states=None,
        ex_filter_folders=None,
        ex_filter_names=None,
        ex_filter_hosts=None,
        ex_filter_clusters=None,
        ex_filter_vms=None,
        ex_filter_datacenters=None,
        ex_filter_resource_pools=None,
        max_properties=20,
    ):
        """
        The ex parameters are search options and must be an array of strings,
        even ex_filter_power_states which can have at most two items but makes
        sense to keep only one ("POWERED_ON" or "POWERED_OFF")
        Keep in mind that this method will return up to 1000 nodes so if your
        network has more, do use the provided filters and call it multiple
        times.
        """

        req = "/rest/vcenter/vm"
        kwargs = {
            "filter.power_states": ex_filter_power_states,
            "filter.folders": ex_filter_folders,
            "filter.names": ex_filter_names,
            "filter.hosts": ex_filter_hosts,
            "filter.clusters": ex_filter_clusters,
            "filter.vms": ex_filter_vms,
            "filter.datacenters": ex_filter_datacenters,
            "filter.resource_pools": ex_filter_resource_pools,
        }
        params = {}

        for param, value in kwargs.items():
            if value:
                params[param] = value
        result = self._request(req, params=params).object["value"]
        vm_ids = [[item["vm"]] for item in result]
        vms = []
        interfaces = self._list_interfaces()

        for vm_id in vm_ids:
            vms.append(self._to_node(vm_id, interfaces))

        return vms

    def async_list_nodes(self):
        """
        In this case filtering is not possible.
        Use this method when the cloud has
        a lot of vms and you want to return them all.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(self._get_all_vms())
        vm_ids = [(item["vm"], item["host"]) for item in result]
        interfaces = self._list_interfaces()

        return loop.run_until_complete(self._list_nodes_async(vm_ids, interfaces))

    async def _list_nodes_async(self, vm_ids, interfaces):
        loop = asyncio.get_event_loop()
        vms = [
            loop.run_in_executor(None, self._to_node, vm_ids[i], interfaces)
            for i in range(len(vm_ids))
        ]

        return await asyncio.gather(*vms)

    async def _get_all_vms(self):
        """
        6.7 doesn't offer any pagination, if we get 1000 vms we will try
        this roundabout  way: First get all the datacenters, for each
        datacenter get the hosts and for each host the vms it has.
        This assumes that datacenters, hosts per datacenter and vms per
        host don't exceed 1000.
        """
        datacenters = self.ex_list_datacenters()
        loop = asyncio.get_event_loop()
        hosts_futures = [
            loop.run_in_executor(
                None,
                functools.partial(self.ex_list_hosts, ex_filter_datacenters=datacenter["id"]),
            )
            for datacenter in datacenters
        ]
        hosts = await asyncio.gather(*hosts_futures)

        vm_resp_futures = [
            loop.run_in_executor(None, functools.partial(self._get_vms_with_host, host))
            for host in itertools.chain(*hosts)
        ]

        vm_resp = await asyncio.gather(*vm_resp_futures)
        # return a flat list

        return [item for vm_list in vm_resp for item in vm_list]

    def _get_vms_with_host(self, host):
        req = "/rest/vcenter/vm"
        host_id = host["host"]
        response = self._request(req, params={"filter.hosts": host_id})
        vms = response.object["value"]

        for vm in vms:
            vm["host"] = host

        return vms

    def list_locations(self, ex_show_hosts_in_drs=True):
        """
        Location in the general sense means any resource that allows for node
        creation. In vSphere's case that usually is a host but if a cluster
        has rds enabled, a cluster can be assigned to create the VM, thus the
        clusters with rds enabled will be added to locations.

        :param ex_show_hosts_in_drs: A DRS cluster schedules automatically
                                     on which host should be placed thus it
                                     may not be desired to show the hosts
                                     in a DRS enabled cluster. Set False to
                                     not show these hosts.
        :type ex_show_hosts_in_drs:  `boolean`
        """
        clusters = self.ex_list_clusters()
        hosts_all = self.ex_list_hosts()
        hosts = []

        if ex_show_hosts_in_drs:
            hosts = hosts_all
        else:
            cluster_filter = [cluster["cluster"] for cluster in clusters]
            filter_hosts = self.ex_list_hosts(ex_filter_clusters=cluster_filter)
            hosts = [host for host in hosts_all if host not in filter_hosts]
        driver = self.connection.driver
        locations = []

        for cluster in clusters:
            if cluster["drs_enabled"]:
                extra = {"type": "cluster", "drs": True, "ha": cluster["ha_enabled"]}
                locations.append(
                    NodeLocation(
                        id=cluster["cluster"],
                        name=cluster["name"],
                        country="",
                        driver=driver,
                        extra=extra,
                    )
                )

        for host in hosts:
            extra = {
                "type": "host",
                "status": host["connection_state"],
                "state": host["power_state"],
            }

            locations.append(
                NodeLocation(
                    id=host["host"],
                    name=host["name"],
                    country="",
                    driver=driver,
                    extra=extra,
                )
            )

        return locations

    def stop_node(self, node):
        if node.state == NodeState.STOPPED:
            return True

        method = "POST"
        req = "/rest/vcenter/vm/{}/power/stop".format(node.id)

        result = self._request(req, method=method)

        return result.status in self.VALID_RESPONSE_CODES

    def start_node(self, node):
        if isinstance(node, str):
            node_id = node
        else:
            if node.state is NodeState.RUNNING:
                return True
            node_id = node.id
        method = "POST"
        req = "/rest/vcenter/vm/{}/power/start".format(node_id)
        result = self._request(req, method=method)

        return result.status in self.VALID_RESPONSE_CODES

    def reboot_node(self, node):
        if node.state is not NodeState.RUNNING:
            return False

        method = "POST"
        req = "/rest/vcenter/vm/{}/power/reset".format(node.id)
        result = self._request(req, method=method)

        return result.status in self.VALID_RESPONSE_CODES

    def destroy_node(self, node):
        # make sure the machine is stopped

        if node.state is not NodeState.STOPPED:
            self.stop_node(node)
        # wait to make sure it stopped
        # in the future this can be made asynchronously
        # for i in range(6):
        #     if node.state is NodeState.STOPPED:
        #         break
        #     time.sleep(10)
        req = "/rest/vcenter/vm/{}".format(node.id)
        resp = self._request(req, method="DELETE")

        return resp.status in self.VALID_RESPONSE_CODES

    def ex_suspend_node(self, node):
        if node.state is not NodeState.RUNNING:
            return False

        method = "POST"
        req = "/rest/vcenter/vm/{}/power/suspend".format(node.id)
        result = self._request(req, method=method)

        return result.status in self.VALID_RESPONSE_CODES

    def _list_interfaces(self):
        request = "/rest/appliance/networking/interfaces"
        response = self._request(request).object["value"]
        interfaces = [
            {
                "name": interface["name"],
                "mac": interface["mac"],
                "status": interface["status"],
                "ip": interface["ipv4"]["address"],
            }
            for interface in response
        ]

        return interfaces

    def _to_node(self, vm_id_host, interfaces):
        """
        id, name, state, public_ips, private_ips,
                driver, size=None, image=None, extra=None, created_at=None)
        """
        vm_id = vm_id_host[0]
        req = "/rest/vcenter/vm/" + vm_id
        vm = self._request(req).object["value"]
        name = vm["name"]
        state = self.NODE_STATE_MAP[vm["power_state"].lower()]

        # IP's
        private_ips = []
        nic_macs = set()

        for nic in vm["nics"]:
            nic_macs.add(nic["value"]["mac_address"])

        for interface in interfaces:
            if interface["mac"] in nic_macs:
                private_ips.append(interface["ip"])
                nic_macs.remove(interface["mac"])

                if len(nic_macs) == 0:
                    break
        public_ips = []  # should default_getaway be the public?

        driver = self.connection.driver

        # size
        total_size = 0
        gb_converter = 1024**3

        for disk in vm["disks"]:
            total_size += int(int(disk["value"]["capacity"] / gb_converter))
        ram = int(vm["memory"]["size_MiB"])
        cpus = int(vm["cpu"]["count"])
        id_to_hash = str(ram) + str(cpus) + str(total_size)
        size_id = hashlib.md5(id_to_hash.encode("utf-8")).hexdigest()  # nosec
        size_name = name + "-size"
        size_extra = {"cpus": cpus}
        size = NodeSize(
            id=size_id,
            name=size_name,
            ram=ram,
            disk=total_size,
            bandwidth=0,
            price=0,
            driver=driver,
            extra=size_extra,
        )

        # image
        image_name = vm["guest_OS"]
        image_id = image_name + "_id"
        image_extra = {"type": "guest_OS"}
        image = NodeImage(id=image_id, name=image_name, driver=driver, extra=image_extra)
        extra = {"snapshots": []}

        if len(vm_id_host) > 1:
            extra["host"] = vm_id_host[1].get("name", "")

        return Node(
            id=vm_id,
            name=name,
            state=state,
            public_ips=public_ips,
            private_ips=private_ips,
            driver=driver,
            size=size,
            image=image,
            extra=extra,
        )

    def ex_list_hosts(
        self,
        ex_filter_folders=None,
        ex_filter_standalone=None,
        ex_filter_hosts=None,
        ex_filter_clusters=None,
        ex_filter_names=None,
        ex_filter_datacenters=None,
        ex_filter_connection_states=None,
    ):
        kwargs = {
            "filter.folders": ex_filter_folders,
            "filter.names": ex_filter_names,
            "filter.hosts": ex_filter_hosts,
            "filter.clusters": ex_filter_clusters,
            "filter.standalone": ex_filter_standalone,
            "filter.datacenters": ex_filter_datacenters,
            "filter.connection_states": ex_filter_connection_states,
        }

        params = {}

        for param, value in kwargs.items():
            if value:
                params[param] = value
        req = "/rest/vcenter/host"
        result = self._request(req, params=params).object["value"]

        return result

    def ex_list_clusters(
        self,
        ex_filter_folders=None,
        ex_filter_names=None,
        ex_filter_datacenters=None,
        ex_filter_clusters=None,
    ):
        kwargs = {
            "filter.folders": ex_filter_folders,
            "filter.names": ex_filter_names,
            "filter.datacenters": ex_filter_datacenters,
            "filter.clusters": ex_filter_clusters,
        }
        params = {}

        for param, value in kwargs.items():
            if value:
                params[param] = value
        req = "/rest/vcenter/cluster"
        result = self._request(req, params=params).object["value"]

        return result

    def ex_list_datacenters(
        self, ex_filter_folders=None, ex_filter_names=None, ex_filter_datacenters=None
    ):
        req = "/rest/vcenter/datacenter"
        kwargs = {
            "filter.folders": ex_filter_folders,
            "filter.names": ex_filter_names,
            "filter.datacenters": ex_filter_datacenters,
        }
        params = {}

        for param, value in kwargs.items():
            if value:
                params[param] = value
        result = self._request(req, params=params)
        to_return = [
            {"name": item["name"], "id": item["datacenter"]} for item in result.object["value"]
        ]

        return to_return

    def ex_list_content_libraries(self):
        req = "/rest/com/vmware/content/library"
        try:
            result = self._request(req).object

            return result["value"]
        except BaseHTTPError:
            return []

    def ex_list_content_library_items(self, library_id):
        req = "/rest/com/vmware/content/library/item"
        params = {"library_id": library_id}
        try:
            result = self._request(req, params=params).object

            return result["value"]
        except BaseHTTPError:
            logger.error(
                "Library was cannot be accessed, "
                " most probably the VCenter service "
                "is stopped"
            )

            return []

    def ex_list_folders(self):
        req = "/rest/vcenter/folder"
        response = self._request(req).object
        folders = response["value"]

        for folder in folders:
            folder["id"] = folder["folder"]

        return folders

    def ex_list_datastores(
        self,
        ex_filter_folders=None,
        ex_filter_names=None,
        ex_filter_datacenters=None,
        ex_filter_types=None,
        ex_filter_datastores=None,
    ):
        req = "/rest/vcenter/datastore"
        kwargs = {
            "filter.folders": ex_filter_folders,
            "filter.names": ex_filter_names,
            "filter.datacenters": ex_filter_datacenters,
            "filter.types": ex_filter_types,
            "filter.datastores": ex_filter_datastores,
        }
        params = {}

        for param, value in kwargs.items():
            if value:
                params[param] = value
        result = self._request(req, params=params).object["value"]

        for datastore in result:
            datastore["id"] = datastore["datastore"]

        return result

    def ex_update_memory(self, node, ram):
        """
        :param ram: The amount of ram in MB.
        :type ram: `str` or `int`
        """

        if isinstance(node, str):
            node_id = node
        else:
            node_id = node.id
        request = "/rest/vcenter/vm/{}/hardware/memory".format(node_id)
        ram = int(ram)

        body = {"spec": {"size_MiB": ram}}
        response = self._request(request, method="PATCH", data=json.dumps(body))

        return response.status in self.VALID_RESPONSE_CODES

    def ex_update_cpu(self, node, cores):
        """
        Assuming 1 Core per socket
        :param cores: Integer or string indicating number of cores
        :type cores: `int` or `str`
        """

        if isinstance(node, str):
            node_id = node
        else:
            node_id = node.id
        request = "/rest/vcenter/vm/{}/hardware/cpu".format(node_id)
        cores = int(cores)
        body = {"spec": {"count": cores}}
        response = self._request(request, method="PATCH", data=json.dumps(body))

        return response.status in self.VALID_RESPONSE_CODES

    def ex_update_capacity(self, node, capacity):
        # Should be added when REST API supports it
        pass

    def ex_add_nic(self, node, network):
        """
        Creates a network adapter that will connect to the specified network
        for the given node. Returns a boolean indicating success or not.
        """

        if isinstance(node, str):
            node_id = node
        else:
            node_id = node.id
        spec = {}
        spec["mac_type"] = "GENERATED"
        spec["backing"] = {}
        spec["backing"]["type"] = "STANDARD_PORTGROUP"
        spec["backing"]["network"] = network
        spec["start_connected"] = True

        data = json.dumps({"spec": spec})
        req = "/rest/vcenter/vm/{}/hardware/ethernet".format(node_id)
        method = "POST"
        resp = self._request(req, method=method, data=data)

        return resp.status

    def _get_library_item(self, item_id):
        req = "/rest/com/vmware/content/library/item/id:{}".format(item_id)
        result = self._request(req).object

        return result["value"]

    def _get_resource_pool(self, host_id=None, cluster_id=None, name=None):
        pms = None

        if host_id:
            pms = {"filter.hosts": host_id}

        if cluster_id:
            pms = {"filter.clusters": cluster_id}

        if name:
            pms = {"filter.names": name}

        if not pms:
            raise ValueError("Either host_id, cluster_id or name must be provided")

        rp_request = "/rest/vcenter/resource-pool"
        resource_pool = self._request(rp_request, params=pms).object

        return resource_pool["value"][0]["resource_pool"]

    def _request(self, req, method="GET", params=None, data=None):
        try:
            result = self.connection.request(req, method=method, params=params, data=data)
        except BaseHTTPError as exc:
            if exc.code == 401:
                self.connection.session_token = None
                self._get_session_token()
                result = self.connection.request(req, method=method, params=params, data=data)
            else:
                raise
        except Exception:
            raise

        return result

    def list_images(self, **kwargs):
        libraries = self.ex_list_content_libraries()
        item_ids = []

        if libraries:
            for library in libraries:
                item_ids.extend(self.ex_list_content_library_items(library))
        items = []

        if item_ids:
            for item_id in item_ids:
                items.append(self._get_library_item(item_id))
        images = []
        names = set()

        if items:
            driver = self.connection.driver

            for item in items:
                names.add(item["name"])
                extra = {"type": item["type"]}

                if item["type"] == "vm-template":
                    capacity = item["size"] // (1024**3)
                    extra["disk_size"] = capacity
                images.append(
                    NodeImage(id=item["id"], name=item["name"], driver=driver, extra=extra)
                )

        if self.driver_soap is None:
            self._get_soap_driver()
        templates_in_hosts = self.driver_soap.list_images()

        for template in templates_in_hosts:
            if template.name not in names:
                images += [template]

        return images

    def ex_list_networks(self):
        request = "/rest/vcenter/network"
        response = self._request(request).object["value"]
        networks = []

        for network in response:
            networks.append(
                VSphereNetwork(
                    id=network["network"],
                    name=network["name"],
                    extra={"type": network["type"]},
                )
            )

        return networks

    def create_node(
        self,
        name,
        image,
        size=None,
        location=None,
        ex_datastore=None,
        ex_disks=None,
        ex_folder=None,
        ex_network=None,
        ex_turned_on=True,
    ):
        """
        Image can be either a vm template , a ovf template or just
        the guest OS.

        ex_folder is necessary if the image is a vm-template, this method
        will attempt to put the VM in a random folder and a warning about it
        will be issued in case the value remains `None`.
        """
        create_request = None
        data = {}

        # image is in the host then need the 6.5 driver

        if image.extra["type"] == "template_6_5":
            kwargs = {}
            kwargs["name"] = name
            kwargs["image"] = image
            kwargs["size"] = size
            kwargs["ex_network"] = ex_network
            kwargs["location"] = location

            for dstore in self.ex_list_datastores():
                if dstore["id"] == ex_datastore:
                    kwargs["ex_datastore"] = dstore["name"]

                    break
            kwargs["folder"] = ex_folder

            if self.driver_soap is None:
                self._get_soap_driver()
            result = self.driver_soap.create_node(**kwargs)

            return result

        # post creation checks
        create_nic = False
        update_memory = False
        update_cpu = False
        create_disk = False
        update_capacity = False

        if image.extra["type"] == "guest_OS":
            spec = {}
            spec["guest_OS"] = image.name
            spec["name"] = name
            spec["placement"] = {}

            if ex_folder is None:
                warn = (
                    "The API(6.7) requires the folder to be given, I will"
                    " place it into a random folder, after creation you "
                    "might find it convenient to move it into a better "
                    "folder."
                )
                warnings.warn(warn)
                folders = self.ex_list_folders()

                for folder in folders:
                    if folder["type"] == "VIRTUAL_MACHINE":
                        ex_folder = folder["folder"]

                if ex_folder is None:
                    msg = "No suitable folder for VMs found, please create one"
                    raise ProviderError(msg, 404)
            spec["placement"]["folder"] = ex_folder

            if location.extra["type"] == "host":
                spec["placement"]["host"] = location.id
            elif location.extra["type"] == "cluster":
                spec["placement"]["cluster"] = location.id
            elif location.extra["type"] == "resource_pool":
                spec["placement"]["resource_pool"] = location.id
            spec["placement"]["datastore"] = ex_datastore
            cpu = size.extra.get("cpu", 1)
            spec["cpu"] = {"count": cpu}
            spec["memory"] = {"size_MiB": size.ram}

            if size.disk:
                disk = {}
                disk["new_vmdk"] = {}
                disk["new_vmdk"]["capacity"] = size.disk * (1024**3)
                spec["disks"] = [disk]

            if ex_network:
                nic = {}
                nic["mac_type"] = "GENERATED"
                nic["backing"] = {}
                nic["backing"]["type"] = "STANDARD_PORTGROUP"
                nic["backing"]["network"] = ex_network
                nic["start_connected"] = True
                spec["nics"] = [nic]
            create_request = "/rest/vcenter/vm"
            data = json.dumps({"spec": spec})

        elif image.extra["type"] == "ovf":
            ovf_request = (
                "/rest/com/vmware/vcenter/ovf/library-item" "/id:{}?~action=filter".format(image.id)
            )
            spec = {}
            spec["target"] = {}

            if location.extra.get("type") == "resource-pool":
                spec["target"]["resource_pool_id"] = location.id
            elif location.extra.get("type") == "host":
                resource_pool = self._get_resource_pool(host_id=location.id)

                if not resource_pool:
                    msg = (
                        "Could not find resource-pool for given location "
                        "(host). Please make sure the location is valid."
                    )
                    raise VSphereException(code="504", message=msg)
                spec["target"]["resource_pool_id"] = resource_pool
                spec["target"]["host_id"] = location.id
            elif location.extra.get("type") == "cluster":
                resource_pool = self._get_resource_pool(cluster_id=location.id)

                if not resource_pool:
                    msg = (
                        "Could not find resource-pool for given location "
                        "(cluster). Please make sure the location "
                        "is valid."
                    )
                    raise VSphereException(code="504", message=msg)
                spec["target"]["resource_pool_id"] = resource_pool
            ovf = self._request(ovf_request, method="POST", data=json.dumps(spec)).object["value"]
            spec["deployment_spec"] = {}
            spec["deployment_spec"]["name"] = name
            # assuming that since you want to make a vm you don't need reminder
            spec["deployment_spec"]["accept_all_EULA"] = True
            # network

            if ex_network and ovf["networks"]:
                spec["deployment_spec"]["network_mappings"] = [
                    {"key": ovf["networks"][0], "value": ex_network}
                ]
            elif not ovf["networks"] and ex_network:
                create_nic = True
            # storage

            if ex_datastore:
                spec["deployment_spec"]["storage_mappings"] = []
                store_map = {"type": "DATASTORE", "datastore_id": ex_datastore}
                spec["deployment_spec"]["storage_mappings"].append(store_map)

            if size and size.ram:
                update_memory = True

            if size and size.extra and size.extra.get("cpu"):
                update_cpu = True

            if size and size.disk:
                # TODO Should update capacity but it is not possible with 6.7
                pass

            if ex_disks:
                create_disk = True

            create_request = (
                "/rest/com/vmware/vcenter/ovf/library-item" "/id:{}?~action=deploy".format(image.id)
            )
            data = json.dumps(
                {"target": spec["target"], "deployment_spec": spec["deployment_spec"]}
            )

        elif image.extra["type"] == "vm-template":
            tp_request = "/rest/vcenter/vm-template/library-items/" + image.id
            template = self._request(tp_request).object["value"]
            spec = {}
            spec["name"] = name

            # storage

            if ex_datastore:
                spec["disk_storage"] = {}
                spec["disk_storage"]["datastore"] = ex_datastore

            # location :: folder,resource group, datacenter, host
            spec["placement"] = {}

            if not ex_folder:
                warn = (
                    "The API(6.7) requires the folder to be given, I will"
                    " place it into a random folder, after creation you "
                    "might find it convenient to move it into a better "
                    "folder."
                )
                warnings.warn(warn)
                folders = self.ex_list_folders()

                for folder in folders:
                    if folder["type"] == "VIRTUAL_MACHINE":
                        ex_folder = folder["folder"]

                if ex_folder is None:
                    msg = "No suitable folder for VMs found, please create one"
                    raise ProviderError(msg, 404)
            spec["placement"]["folder"] = ex_folder

            if location.extra["type"] == "host":
                spec["placement"]["host"] = location.id
            elif location.extra["type"] == "cluster":
                spec["placement"]["cluster"] = location.id
            # network changes the network to existing nics if
            # there are no adapters
            # in the template then we will make on in the vm
            # after the creation finishes
            # only one network atm??
            spec["hardware_customization"] = {}

            if ex_network:
                nics = template["nics"]

                if len(nics) > 0:
                    nic = nics[0]
                    spec["hardware_customization"]["nics"] = [
                        {"key": nic["key"], "value": {"network": ex_network}}
                    ]
                else:
                    create_nic = True
            spec["powered_on"] = False
            # hardware

            if size:
                if size.ram:
                    spec["hardware_customization"]["memory_update"] = {"memory": int(size.ram)}

                if size.extra.get("cpu"):
                    spec["hardware_customization"]["cpu_update"] = {"num_cpus": size.extra["cpu"]}

                if size.disk:
                    if not len(template["disks"]) > 0:
                        create_disk = True
                    else:
                        capacity = size.disk * 1024 * 1024 * 1024
                        dsk = template["disks"][0]["key"]

                        if template["disks"][0]["value"]["capacity"] < capacity:
                            update = {"capacity": capacity}
                            spec["hardware_customization"]["disks_to_update"] = [
                                {"key": dsk, "value": update}
                            ]

            create_request = "/rest/vcenter/vm-template/library-items/" "{}/?action=deploy".format(
                image.id
            )
            data = json.dumps({"spec": spec})

        if not create_request:
            raise ValueError("Missing create_request")

        # deploy the node
        result = self._request(create_request, method="POST", data=data)
        # wait until the node is up and then add extra config
        node_id = result.object["value"]

        if image.extra["type"] == "ovf":
            node_id = node_id["resource_id"]["id"]

        node = self.list_nodes(ex_filter_vms=node_id)[0]

        if create_nic:
            self.ex_add_nic(node, ex_network)

        if update_memory:
            self.ex_update_memory(node, size.ram)

        if update_cpu:
            self.ex_update_cpu(node, size.extra["cpu"])

        if create_disk:
            pass  # until volumes are added

        if update_capacity:
            pass  # until API method is added

        if ex_turned_on:
            self.start_node(node)

        return node

    # TODO As soon as snapshot support gets added to the REST api
    # these methods should be rewritten with REST api calls
    def ex_list_snapshots(self, node):
        """
        List node snapshots
        """

        if self.driver_soap is None:
            self._get_soap_driver()

        return self.driver_soap.ex_list_snapshots(node)

    def ex_create_snapshot(
        self, node, snapshot_name, description="", dump_memory=False, quiesce=False
    ):
        """
        Create node snapshot
        """

        if self.driver_soap is None:
            self._get_soap_driver()

        return self.driver_soap.ex_create_snapshot(
            node,
            snapshot_name,
            description=description,
            dump_memory=dump_memory,
            quiesce=False,
        )

    def ex_remove_snapshot(self, node, snapshot_name=None, remove_children=True):
        """
        Remove a snapshot from node.
        If snapshot_name is not defined remove the last one.
        """

        if self.driver_soap is None:
            self._get_soap_driver()

        return self.driver_soap.ex_remove_snapshot(
            node, snapshot_name=snapshot_name, remove_children=remove_children
        )

    def ex_revert_to_snapshot(self, node, snapshot_name=None):
        """
        Revert node to a specific snapshot.
        If snapshot_name is not defined revert to the last one.
        """

        if self.driver_soap is None:
            self._get_soap_driver()

        return self.driver_soap.ex_revert_to_snapshot(node, snapshot_name=snapshot_name)

    def ex_open_console(self, vm_id):
        if self.driver_soap is None:
            self._get_soap_driver()

        return self.driver_soap.ex_open_console(vm_id)
